# -*- coding: utf-8 -*-

import time
import logging

import numpy as np
import numpy.matlib

import pyhrf

from pyhrf.ndarray import xndarray, stack_cuboids
from pyhrf import xmlio
from pyhrf.tools import array_summary, get_2Dtable_string


logger = logging.getLogger(__name__)


class DuplicateVariableException(Exception):
    def __init__(self, vName, v):
        self.variableName = vName
        self.variableObject = v

    def __str__(self):
        return 'vname:' + str(self.variableName)


class VariableTypeException(Exception):
    def __init__(self, vClass, vName, v):
        self.variableName = vName
        self.variableObject = v
        self.variableClass = vClass


class GibbsSampler:

    """
    Generic class of a Gibbs sampler with gathers common operations for any
    gibbs sampling: variable initialisation, observables updates (posterior mean),
    outputs ...
    """

    def __init__(self, variables, nbIt, smplHistoryPace=-1,
                 obsHistoryPace=-1, nbSweeps=None,
                 callbackObj=None, randomSeed=None, globalObsHistoryPace=-1,
                 check_ftval=None, output_fit=False):
        """
        Initialize a new GibbsSampler object.
        @param variables: contains all instances of type C{GibbsSamplerVariable}
        involved in sampling.
        @param dataInput: data to fit
        @param nbIt: is the number of maximum steps to terminate the sampling
        @param callbackObj: is an object which function L{callback()} is called
        after each sampling iteration (typically for plotting/report purpose.
        See C{GSDefaultCallbackHandler}.
        """

        # TODO change to Exceptions ...
        # Check for type consistency :
        if type(variables) != type(list()):
            logger.error('Error : given variables of type : \'%s\' '
                         '-> need a list', type(variables))
            return
        self.variablesMapping = {}
        for var in variables:
            # if var is not an instance of GibbsSamplerVariable
            if not isinstance(var, GibbsSamplerVariable):
                # TODO : exception ...
                logger.error('!Err : variable(%s) of wrong type -> '
                             'need a derivation of GibbsSamplerVariable',
                             var.__class__)
                raise VariableTypeException(var.__class__, var.name, var)
            else:
                if not self.variablesMapping.has_key(var.name):
                    self.variablesMapping[var.name] = var
                else:
                    raise DuplicateVariableException(var.name, var)
        self.variables = variables  # variables involved in sampling

        if callbackObj is None:
            self.callbacker = GSDefaultCallbackHandler()
        else:
            self.callbacker = callbackObj

        self.nbSweeps = nbSweeps
        self.nbIterations = nbIt
        self.output_fit = output_fit
        self.smplHistoryPace = smplHistoryPace
        self.obsHistoryPace = obsHistoryPace
        self.globalObsHistoryPace = globalObsHistoryPace
        self.globalObsHistoryIts = None
        self.globalObsHistoryTiming = None

        self.tSamplingOnly = 0
        self.analysis_duration = 0
        self.tVars = np.zeros(len(variables), dtype=float)

        for v in variables:
            v.registerNbIterations(nbIt)
            v.setSamplerEngine(self)

        self.sharedData = None
        self.randomSeed = randomSeed

        self.check_ftval = check_ftval

    def set_nb_iterations(self, n):
        prev_its = self.nbIterations
        self.nbIterations = n

        logger.debug('set_nb_iterations to %d', n)
        logger.debug('prev nb sweep %d', self.nbSweeps)

        if self.nbSweeps > 0:
            self.nbSweeps = int(np.ceil(self.nbSweeps * (n * 1. / prev_its)))

        logger.debug('new nb sweep %d', self.nbSweeps)

        if self.smplHistoryPace > 0:
            self.smplHistoryPace = int(np.ceil(self.smplHistoryPace *
                                               (n * 1. / prev_its)))

        if self.obsHistoryPace > 0:
            self.obsHistoryPace = int(np.ceil(self.obsHistoryPace *
                                              (n * 1. / prev_its)))

    def linkToData(self, dataInput):
        #-> set input data
        self.dataInput = dataInput
        self.fit = None  # TODO init fit according to input signal shape
        for v in self.variables:
            logger.info('%s -> linking to data', v.name)
            v.linkToData(dataInput)
            if (v.axes_names is not None) and ('condition' in v.axes_names):
                v.axes_domains['condition'] = self.dataInput.cNames

    def regVarsInPipeline(self):

        if self.sharedData == None:
            raise Exception("sharedData isn\'t initialized !")

        rules = []
        for v in self.variables:
            rules.append({'label': v.name, 'ref': v})

        self.sharedData.addDependencies(rules)

    def get_variable(self, label):
        return self.variablesMapping[label]

    def iterate_sampling(self):
        it = 0
        while it < self.nbIterations and not self.stop_criterion(it):
            yield it
            it += 1

    def stop_criterion(self, it):
        return False

    def runSampling(self, atomData=None):
        # np.seterr(all='raise')
        """
        Launch a complete sampling process by calling the function
        L{GibbsSamplerVariable.sampleNext()} of each variable. Call the callback
        function after each iteration. Measure time elapsed and store it in
        L{tSamplinOnly} and L{analysis_duration}
        """
        if self.randomSeed is not None:
            logger.info('setting random seed: %s', str(self.randomSeed))
            np.random.seed(self.randomSeed)
        else:
            logger.info('No random seed specified ')

        #self.itAxis = []
        if self.nbSweeps is None:
            self.nbSweeps = self.nbIterations / 3

        logger.info('Nb sweeps: %d', self.nbSweeps)
        logger.info('obs hist pace: %d', self.obsHistoryPace)
        logger.info('smpl hist pace: %d', self.obsHistoryPace)

        for v in self.variables:
            logger.info('Init of ' + v.name)
            logger.info(v.get_summary())
            v.checkAndSetInitValue(self.variables)
            if type(v.currentValue) == numpy.ndarray:
                logger.debug(v.currentValue)
            else:
                logger.debug(v.currentValue)

        # Warming up variables to sample (precalculations):
        for v in self.variables:
            if v.useTrueValue:
                if v.trueValue is None:
                    raise Exception("No true value defined for %s" % v.name)

            logger.info('Warming of ' + v.name)
            v.samplingWarmUp(self.variables)
            v.initObservables()
        self.initGlobalObservables()

        # init for time measures :
        tGlobIni = time.time()
        tIni = time.time()

        # main loop over variables to sample :
        nbIt2EstimTime = max(1, min(10, int(self.nbIterations * 0.1)))
        tLoopShown = False
        logger.info('Starting sampling (%d it)...', self.nbIterations)
        try:
            lhrf = self.get_variable('hrf').hrfLength
        except KeyError:
            lhrf = self.get_variable('brf').hrfLength

        logger.info('Data dims : nbvox=%d, ny=%d, nbCond=%d, nb coeffs hrf=%d',
                    self.dataInput.nbVoxels, self.dataInput.ny,
                    self.dataInput.nbConditions, lhrf)

        tLoopIni = time.time()
        for v in self.variables:
            if self.smplHistoryPace != -1:
                logger.info('Saving init value of %s', v.name)
                v.saveCurrentValue(-1)

        rerror = np.array([])
        loglkhd = np.array([])

        for it in self.iterate_sampling():
            iv = 0

            for v in self.variables:
                tIniv = time.time()

                if not v.sampleFlag:
                    logger.info('(it %d) %s not sampled', it, v.name)
                else:
                    logger.info('(it %d) Sampling of %s ...', it, v.name)
                v.sampleNext(self.variables)

                logger.debug('current value of %s:', v.name)
                logger.debug(v.currentValue)
                if v.sampleFlag:
                    logger.info('(it %d) Sampling of %s done !', it, v.name)
                self.tVars[iv] += time.time() - tIniv
                iv += 1

            for v in self.variables:

                v.record_trajectories(it)

                if self.smplHistoryPace != -1 and (it % self.smplHistoryPace) == 0:
                    logger.info('(it %d) saving sample of %s)', it, v.name)
                    v.saveCurrentValue(it)

                # Update observables after burn-in period :
                if it >= self.nbSweeps:
                    v.updateObsersables()
                    if self.obsHistoryPace != -1 and (it % self.obsHistoryPace) == 0:
                        logger.info(
                            '(it %d) saving observables of %s', it, v.name)
                        v.saveObservables(it)

            if it >= self.nbSweeps:
                self.updateGlobalObservables()
                if self.globalObsHistoryPace != -1 and \
                        (it % self.globalObsHistoryPace) == 0:
                    self.saveGlobalObservables(it)

            # if pyhrf.verbose.verbosity >= 6:
            #     try:
            #         f = self.computeFit()
            #         err = np.sqrt((self.dataInput.varMBY - f)**2)
            #         pyhrf.verbose(6, 'Errors:')
            #         pyhrf.verbose.printNdarray(6, err)
            #     except NotImplementedError:
            #         pass

            # Save jde_fit()
            #self.jde_fit_vec = np.append(self.jde_fit_vec, self.computeFit())

            # Compute error measures
            try:
                bold = atomData.bold

                # relative reconstruction error
                jde_fit = self.computeFit()
                r = bold - jde_fit
                rec_error_j = np.sum(r ** 2, 0)
                bold2 = np.sum(bold ** 2, 0)
                #rec_error = np.mean(rec_error_j/bold2)   # Univariate analysis
                rec_error = np.mean(rec_error_j) / np.mean(bold2)
                rerror = np.append(rerror, rec_error)

                # Loglikelihood
                var_noise = self.get_variable('noise_var').currentValue
                loglh = 0
                N = r.shape[0]
                J = r.shape[1]
                for j in np.arange(0, J):
                    loglh -= (np.log(np.abs(2*np.pi*var_noise[j]*N)) + \
                        np.dot(r[:,j].T,r[:,j])/var_noise[j] / 2)
                loglkhd = np.append(loglkhd, loglh)

            except AttributeError:
                pass


            # Some verbose about online profiling :
            now = time.time()
            self.loop_timing = (now - tLoopIni)

            if not tLoopShown and it >= nbIt2EstimTime - 1:
                tLoop = self.loop_timing / nbIt2EstimTime
                tSampling = tLoop * (self.nbIterations - nbIt2EstimTime)
                logger.info('Time of loop : %1.2fs => total time ~%1.2fs',
                            tLoop, tSampling)
                logger.info(
                    'Now is: %s', time.strftime('%c', time.localtime()))
                logger.info('Sampling should end at : %s',
                            time.strftime('%c', time.localtime(now + tSampling)))
                tLoopShown = True

            # measure the time spent in sampling only :
            tFin = time.time()
            self.tSamplingOnly += tFin - tIni

            # launch callback function after each sample step :
            logger.info('calling callback ...')
            self.callbacker(it, self.variables, self)
            tIni = time.time()
        self.final_iteration = it
        logger.info('##- Sampling done, final iteration=%d -##',
                    self.final_iteration)

        try:
            bold = atomData.bold
            Q, J = bold.shape

            #BIC
            self.converror = rerror
            self.loglikelihood = loglkhd
            try:
                hrf = self.get_variable('brf').currentValue
            except KeyError:
                hrf = self.get_variable('hrf').currentValue
            if len(hrf.shape) > 1:
                M = hrf.shape[1]
            else:
                M = 1
            D = hrf.shape[0]
            p = 2 * M * J + 2 * (D - 1) + J * Q + J
            n = N * J
            self.bic = loglh + p / 2 * np.log(n)
        except AttributeError:
            pass

        # Finalizing everything :
        for v in self.variables:
            if v.finalValue is None:
                v.setFinalValue()

        for v in self.variables:
            logger.info(v.name + '-> finalizing sampling ...')
            v.finalizeSampling()
            logger.info(v.get_final_summary())

        for v in self.variables:
            logger.info(v.name + '-> cleaning observables ...')
            v.cleanObservables()

        logger.info('Finalizing overall sampling ...')

        self.finalizeSampling()
        #outputs = self.getGlobalOutputs()
        # measure time for sampling and callback :
        self.analysis_duration = time.time() - tGlobIni
        #print self.getTinyProfile()

        return

    def finalizeSampling(self):
        pass

    def getOutputs(self):
        outputs = {}
        logger.info('get output of sampled variables ...')
        for v in self.variables:
            logger.debug('get outputs of %s', v.name)
            outputs.update(v.getOutputs())

        logger.info('get output of global observables ...')
        outputs.update(self.getGlobalOutputs())

        #tp = time.time()
        d = {'parcel_size': np.array([self.dataInput.nbVoxels])}
        outputs['analysis_duration'] = xndarray(\
                                            np.array([self.analysis_duration]),
                                            axes_names=['parcel_size'],
                                            axes_domains=d)
        try:
            outputs['conv_error'] = xndarray(np.array(self.converror))
            outputs['loglikelihood'] = xndarray(np.array([self.loglikelihood]))
            outputs['bic'] = xndarray(np.array([self.bic]),
                                                axes_names = ['parcel_size'],
                                                axes_domains = d)
        except AttributeError:
            pass

        # for on, o in outputs.iteritems():
        #     #print 'on:', o
        #     yield (on,o)
        return outputs

    def getShortProfile(self):
        prof = ''
        prof += '---------------------\n'
        prof += '|| Short profile   ||\n'
        prof += '---------------------\n'
        prof += '|| Sampling done !\n'  # TODO: add option for printing report on/off
        prof += '|| time spent in sampling : '+ str(self.tSamplingOnly) + ' sec\n'
        prof += '|| time spent in sampling and callbacks : '+ \
                str(self.analysis_duration) + ' sec\n'
        prof += '|| time spent in sampling for each variable : \n'
        for iv in xrange(len(self.variables)):
            v = self.variables[iv]
            prof += '|| ' + v.name + ' -> ' + str(self.tVars[iv]) + ' sec\n'
        return prof

    def getTinyProfile(self):
        return 'Total time spent : ' + str(self.analysis_duration)

    def computeFit(self):
        raise NotImplementedError

    def getFitAxes(self):
        raise NotImplementedError

    def initGlobalObservables(self):
        pass

    def updateGlobalObservables(self):
        pass

    def saveGlobalObservables(self, it):
        if self.globalObsHistoryIts is None:
            self.globalObsHistoryIts = []
        if self.globalObsHistoryTiming is None:
            self.globalObsHistoryTiming = []

        self.globalObsHistoryIts.append(it)
        self.globalObsHistoryTiming.append(self.loop_timing)

    def getGlobalOutputs(self):
        outputs = {}
        axes_domains = {
            'time': np.arange(self.dataInput.ny) * self.dataInput.tr}
        if pyhrf.__usemode__ == pyhrf.DEVEL:
            # output of design matrix:
            dMat = np.zeros_like(self.dataInput.varX[0, :, :])
            for ic, vx in enumerate(self.dataInput.varX):
                dMat += vx * (ic + 1)

            outputs['matX'] = xndarray(dMat, axes_names=['time', 'P'],
                                       axes_domains=axes_domains,
                                       value_label='value')

            ad = axes_domains.copy()
            ad['condition'] = self.dataInput.cNames
            outputs['varX'] = xndarray(self.dataInput.varX.astype(np.int8),
                                       axes_names=['condition', 'time', 'P'],
                                       axes_domains=ad,
                                       value_label='value')
        if self.output_fit:
            try:
                fit = self.computeFit()
                if self.dataInput.varMBY.ndim == 2:
                    axes_names = ['time', 'voxel']
                else:  # multisession
                    axes_names = ['session', 'time', 'voxel']
                bold = xndarray(self.dataInput.varMBY.astype(np.float32),
                                axes_names=axes_names,
                                axes_domains=axes_domains,
                                value_label='BOLD')

                # TODO: outputs of paradigm
                # outputs of onsets, per condition:
                # get binary sequence sampled at TR
                # build a xndarray from it
                # build time axis values

                cfit = xndarray(fit.astype(np.float32),
                                axes_names=axes_names,
                                axes_domains=axes_domains,
                                value_label='BOLD')
                # if self.dataInput.simulData is not None:

                    # s = xndarray(self.dataInput.simulData.stimInduced,
                               # axes_names=axes_names,
                               # axes_domains=axes_domains,
                               # value_label='BOLD')

                    # outputs['fit'] = stack_cuboids([s,cfit], 'type',
                                                   # ['simu', 'fit'])
                # else:
                outputs['bold_fit'] = stack_cuboids([bold,cfit],
                                                    'stype', ['bold', 'fit'])

                # e = np.sqrt((fit.astype(np.float32) - \
                #             self.dataInput.varMBY.astype(np.float32))**2)
                # outputs['error'] = xndarray(e, axes_names=axes_names,
                #                            axes_domains=axes_domains,
                #                            value_label='Error')
                # outputs['rmse'] = xndarray(e.mean(0), axes_names=['voxel'],
                #                            value_label='Rmse')

            except NotImplementedError:
                print 'Compute fit not implemented !'
                pass

        return outputs

    # def initGlobalObservablesOutputs(self, outputs, nbROI):
        # if self.fit is not None:
        #fitAxes = 'iteration' + self.getFitAxes()
        #outputs['fit_history'] = None

    # def fillGlobalObservablesOutputs(self, outputs, iROI):
        # pass
        # if len(self.fitHistory) > 0:


class Trajectory:

    """ Keep track of a numpy array that is modified _inplace_ iteratively
    """

    def __init__(self, variable, axes_names, axes_domains, history_pace,
                 history_start, max_iterations, first_saved_iteration=-1):
        """
        Args:
            *variable* is a numpy array that has to be modified _inplace_ else
            the reference to it is lost.
            *axes_names* is a list of string for the axes of *variable*.
            *axes_domains* is a dict of numpy arrays for the domains of the axes.
            *history_pace* is the pace at which each value of *variable* is saved.
            *history_start* is the iteration when to start saving values of
                            *variable*
            *max_iterations* is the total maximal number of iterations
                           (for allocating memory)
            *first_saved_iteration* sets the integer value of the first
                           saved iteration. -1 is useful to mark the init
                           value which is prior to the 1st iteration (0).
        """
        self.variable = variable
        self.axes_names = axes_names
        self.axes_domains = axes_domains
        self.hist_pace = history_pace
        self.hist_start = history_start

        if history_pace > 0:
            nsamples_max = max_iterations / history_pace - history_start
        else:
            nsamples_max = 0

        if first_saved_iteration == -1:
            nsamples_max += 1  # +1 because of init

        self.history = np.zeros((nsamples_max,) + variable.shape,
                                dtype=variable.dtype)

        self.sample_count = 0
        self.saved_iterations = []
        if first_saved_iteration == -1:
            #-1 is initialisation.
            self.history[0] = self.variable[:]
            self.saved_iterations.append(-1)
            self.sample_count = 1

    def record(self, iteration):
        """
        Increment the history saving.
        """
        if self.hist_pace > 0 and iteration >= self.hist_start and \
                (iteration % self.hist_pace) == 0:
            self.history[self.sample_count] = self.variable[:]
            self.saved_iterations.append(iteration)
            self.sample_count += 1

    def get_last(self):
        """ Return the last saved element """
        return self.history[self.sample_count - 1]

    def to_cuboid(self):
        """ Pack the current trajectory in a xndarray
        """

        if self.axes_names is not None:
            an = ['iteration'] + self.axes_names
        else:
            an = ['iteration'] + \
                ['axis_%02d' % i for i in xrange(self.variable.ndim)]

        if self.axes_domains is not None:
            ad = self.axes_domains.copy()
        else:
            ad = {}
        ad['iteration'] = np.array(self.saved_iterations, dtype=int)
        c = xndarray(self.history, axes_names=an, axes_domains=ad)
        return c


class GibbsSamplerVariable:

    # TODO: make GibbsSamplerVariable be XMLParamDrivenClass
    # to factorize common parametrisation

    def __init__(self, name, valIni=None, trueVal=None, sampleFlag=1,
                 useTrueValue=False, axes_names=None, axes_domains=None,
                 value_label='value'):
        """
        #TODO : comment


        """
        self.axes_domains = {} if axes_domains is None else axes_domains
        self.axes_names = axes_names
        logger.info('Init of GibbsSamplerVariable %s with axes: %s', name,
                    str(axes_names))
        self.value_label = value_label
        self.useTrueValue = useTrueValue
        self.name = name

        if valIni is not None and not isinstance(valIni, np.ndarray):
            raise Exception('Init value of variable %s should be a numpy array '
                            '(if scalar variable then it should encapsulated in'
                            'an array)' % self.name)
        if trueVal is not None and not isinstance(trueVal, np.ndarray):
            raise Exception('true value of variable %s should be a numpy array '
                            '(if scalar variable then it should encapsulated in'
                            'an array)' % self.name)

        if self.useTrueValue:
            self.currentValue = trueVal
        else:
            self.currentValue = valIni  # if None -> must be
            # generated later

        self.trueValue = trueVal  # should only be defined with linkToData (?)
        self.sampleFlag = sampleFlag
        self.sampleNext = self.chooseSampleNext(sampleFlag)
        self.finalValue = None

        self.smplHistory = None
        self.smplHistoryIts = []
        self.meanHistory = None
        self.errorHistory = None
        self.obsHistoryIts = []

        # Used to save history of samples:
        self.tracked_quantities = {}

    def chooseSampleNext(self, flag):
        if flag:
            return self.sampleNextInternal
        else:
            return self.sampleNextAlt

    def get_variable(self, label):
        """
        Return a sibling GibbsSamplerVariable
        """
        return self.samplerEngine.get_variable(label)

    def __setstate__(self, d):
        d['sampleNext'] = self.chooseSampleNext(d['sampleFlag'])
        self.__dict__ = d

    def __getstate__(self):  # use for compatibilty with pickles which doesn't
                            # support instance variable methods
        # copy the __dict__ so that further changes
        # do not affect the current instance
        d = dict(self.__dict__)
        # remove the closure that cannot be pickled
        if d.has_key('sampleNext'):
            del d['sampleNext']
        # return state to be pickled
        return d

    def get_summary(self):
        s = self.name + ' summary:'
        s += '  sample flag:' + ['Off', 'On'][self.sampleFlag]
        s += '  use true value:' + str(self.useTrueValue)
        return s

    def get_string_value(self, v):
        if v.size == 1:
            return '%1.4f' % v[0]
        if self.axes_names is not None and 'condition' == self.axes_names[0]:
            cNames = self.dataInput.cNames
            if v.ndim == 1:
                return get_2Dtable_string(v, cNames, ['Value'])
            else:
                # print 'cNames:', cNames
                # print 'v.shape:', v.shape
                v = v[:len(cNames), :]
                return get_2Dtable_string(v.reshape(v.shape[0], 1, -1),
                                          cNames, ['Value'])
        else:
            return array_summary(v)

    def get_final_summary(self):
        s = self.name + ' sampling report: \n'
        sv = self.get_string_value(self.get_final_value())
        if '\n' in sv:
            s += ' - final value:\n' + sv
        else:
            s += ' - final value: ' + sv + '\n'
        if self.trueValue is not None:
            # TODO: handle condition matching
            self.trueValue = np.array(self.trueValue)
            sv = self.get_string_value(self.trueValue)
            if '\n' in sv:
                s += ' - true value:\n' + sv
            else:
                s += ' - true value: ' + sv + '\n'
            if self.trueValue.shape == self.finalValue.shape:
                err = abs(self.trueValue - self.finalValue)
                sv = self.get_string_value(err)
                if '\n' in sv:
                    s += ' - error:\n' + sv
                else:
                    s += ' - error: ' + sv + '\n'
        return s

    def _track_quantity(self, q, name, axes_names, axes_domains,
                        history_pace, hist_start):
        if not self.tracked_quantities.has_key(name):
            max_its = self.samplerEngine.nbIterations
            trajectory = Trajectory(q, axes_names, axes_domains,
                                    history_pace, hist_start, max_its)
            self.tracked_quantities[name] = trajectory
        else:
            raise Exception('Quantity %s already tracked' % name)

    def track_sampled_quantity(self, q, name, axes_names=None, axes_domains=None,
                               history_pace=None):
        if history_pace is None:
            history_pace = self.samplerEngine.smplHistoryPace

        self._track_quantity(q, name, axes_names, axes_domains,
                             history_pace, hist_start=0)

    def track_obs_quantity(self, q, name, axes_names=None, axes_domains=None,
                           history_pace=None):

        if history_pace is None:
            history_pace = self.samplerEngine.obsHistoryPace

        burnin = self.samplerEngine.nbSweeps
        self._track_quantity(q, name, axes_domains, axes_domains,
                             history_pace, hist_start=burnin)

    def record_trajectories(self, it):
        for q in self.tracked_quantities.values():
            q.record(it)

    def initObservables(self):

        if not np.isscalar(self.currentValue):
            self.cumul = np.zeros(self.currentValue.shape, dtype=np.float64)
            #self.cumul2 = np.zeros(self.currentValue.shape, dtype=np.float64)
            self.cumul3 = np.zeros(self.currentValue.shape, dtype=np.float64)
        else:
            self.cumul = 0.0
            #self.cumul2 = 0.0
            self.cumul3 = 0.0

        self.nbItObservables = 0

    def updateObsersables(self):
        self.nbItObservables += 1
        logger.debug('Generic update Observables for var %s, it=%d ...',
                     self.name, self.nbItObservables)
        logger.debug('CurrentValue:')
        logger.debug(self.currentValue)

        self.cumul += self.currentValue
        #self.cumul2 += self.currentValue**2

        logger.debug('Cumul:')
        logger.debug(self.cumul)

        self.mean = self.cumul / self.nbItObservables

        # Another Computing of error to avoid negative value when cumul is < 1
        self.cumul3 += (self.currentValue - self.mean) ** 2
        logger.debug('Cumul3:')
        logger.debug(self.cumul3)

        logger.debug('Mean')
        logger.debug(self.mean)

        if self.nbItObservables < 2:
            self.error = np.zeros_like(self.cumul3) + 1e-6
        else:
            self.error = self.cumul3 / self.nbItObservables

        tol = 1e-10
        neg_close_to_zero = np.where(np.bitwise_and(self.error < 0,
                                                    np.abs(self.error) < tol))
        self.error[neg_close_to_zero] = tol

        logger.debug('Error:')
        logger.debug(self.error)

        if (self.error < 0.).any():
            raise Exception('neg error on variable %s' % self.name)

    def saveObservables(self, it):

        self.obsHistoryIts.append(it)
        if self.meanHistory is not None:
            self.meanHistory = np.concatenate((self.meanHistory,
                                               [self.mean]))
        else:
            self.meanHistory = np.array([self.mean.copy()])

        if self.errorHistory is not None:
            self.errorHistory = np.concatenate((self.errorHistory,
                                                [self.error]))
        else:
            self.errorHistory = np.array([self.error.copy()])

    def saveCurrentValue(self, it):
        self.smplHistoryIts.append(it)
        if self.smplHistory is not None:
            self.smplHistory = np.concatenate((self.smplHistory,
                                               [self.currentValue]))
        else:
            self.smplHistory = np.array([self.currentValue.copy()])

    def roiMapped(self):
        logger.debug('roiMapped ?')
        logger.debug(' -> self.axes_names : %s', str(self.axes_names))
        return False if (self.axes_names == None) else ('voxel' in self.axes_names)

    def manageMappingInit(self, shape, axes_names):
        tans = self.dataInput.voxelMapping.getTargetAxesNames()
        i = axes_names.index('voxel')
        axes_names = axes_names[:i] + tans + axes_names[i + 1:]
        shape = shape[:i] + self.dataInput.finalShape + shape[i + 1:]
        logger.debug('manageMappingInit returns :')
        logger.debug(' -> sh: %s, axes_names: %s', str(shape), str(axes_names))
        return shape, axes_names

    def getOutputs(self):

        # Backward compatibilty with older results
        if hasattr(self, 'axesNames'):
            self.axes_names = self.axesNames
        if hasattr(self, 'axesDomains'):
            self.axes_domains = self.axesDomains
        if hasattr(self, 'valueLabel'):
            self.value_label = self.valueLabel

        outputs = {}
        if self.axes_names is None:
            # print ' "%s" -> no axes_names defined' %self.name
            sh = (1,) if np.isscalar(
                self.finalValue) else self.finalValue.shape
            #an = ['axis%d'%i for i in xrange(self.finalValue.ndim)]
            an = ['axis%d' % i for i in xrange(len(sh))]
        else:
            an = self.axes_names

        if self.meanHistory is not None:
            outName = self.name + '_pm_history'
            if hasattr(self, 'obsHistoryIts'):
                axes_domains = {'iteration': self.obsHistoryIts}
            else:
                axes_domains = {}
            axes_domains.update(self.axes_domains)

            axes_names = ['iteration'] + an
            outputs[outName] = xndarray(self.meanHistory,
                                        axes_names=axes_names,
                                        axes_domains=axes_domains,
                                        value_label=self.value_label)

        if hasattr(self, 'smplHistory') and self.smplHistory is not None:
            axes_names = ['iteration'] + an
            outName = self.name + '_smpl_history'
            if hasattr(self, 'smplHistoryIts'):
                axes_domains = {'iteration': self.smplHistoryIts}
            else:
                axes_domains = {}
            axes_domains.update(self.axes_domains)
            outputs[outName] = xndarray(self.smplHistory,
                                        axes_names=axes_names,
                                        axes_domains=axes_domains,
                                        value_label=self.value_label)

            if hasattr(self, 'autocorrelation'):
                outName = self.name + '_smpl_autocorr'
                axes_names = ['lag'] + an
                outputs[outName] = xndarray(self.autocorrelation,
                                            axes_names=axes_names,
                                            axes_domains=self.axes_domains,
                                            value_label='acorr')

                outName = self.name + '_smpl_autocorr_test'
                outputs[outName] = xndarray(self.autocorrelation_test,
                                            axes_names=axes_names,
                                            axes_domains=self.axes_domains,
                                            value_label='acorr')

                outName = self.name + '_smpl_autocorr_pval'
                outputs[outName] = xndarray(self.autocorrelation_pvalue,
                                            axes_names=axes_names,
                                            axes_domains=self.axes_domains,
                                            value_label='pvalue')

                outName = self.name + '_smpl_autocorr_thresh'
                outputs[outName] = xndarray(np.array([self.autocorrelation_thresh]),
                                            value_label='acorr')

            if hasattr(self, 'median'):
                outName = self.name + '_post_median'
                outputs[outName] = xndarray(self.median,
                                            axes_names=self.axes_names,
                                            axes_domains=self.axes_domains,
                                            value_label='median')

        logger.info('%s final value:', self.name)
        logger.info(self.finalValue)
        if 1 and hasattr(self, 'error'):
            err = self.error ** .5
        else:
            err = None

        c = xndarray(self.get_final_value().astype(np.float32),
                     axes_names=self.axes_names,
                     axes_domains=self.axes_domains,
                     value_label=self.value_label)

        if self.trueValue is not None:
            c_true = xndarray(np.array(self.get_true_value()),
                              axes_names=self.axes_names,
                              axes_domains=self.axes_domains,
                              value_label=self.value_label)

            c = stack_cuboids([c, c_true], axis='type', domain=['estim', 'true'],
                              axis_pos='last')

        outputs[self.name + '_pm'] = c

        if ((err is not None) or ((err.size == 1) and (err != 0))):
            c_err = xndarray(self.error.astype(np.float32),
                             axes_names=self.axes_names,
                             axes_domains=self.axes_domains,
                             value_label=self.value_label)
            # c = stack_cuboids([c,c_err], axis='error', domain=['value','std'],
            #                   axis_pos='last')
            outputs[self.name + '_mcmc_var'] = c_err

        if hasattr(self, 'tracked_quantities'):
            for qname, q in self.tracked_quantities.iteritems():
                outputs[qname] = q.to_cuboid()

        if len(self.report_check_ft_val) > 0:
            r = self.report_check_ft_val
            outputs[self.name + '_abs_err'] = xndarray(r['abs_error'],
                                                       axes_names=self.axes_names,
                                                       axes_domains=self.axes_domains)

            outputs[self.name + '_rel_err'] = xndarray(r['rel_error'],
                                                       axes_names=self.axes_names,
                                                       axes_domains=self.axes_domains)
            on = self.name + '_inaccuracies'
            an = r['accuracy'][0]
            ad = {}
            if an is not None:
                ad = dict((a, self.axes_domains[a])
                          for a in an if self.axes_domains.has_key(a))
            inacc = np.bitwise_not(r['accuracy'][1]).astype(np.int8)
            outputs[on] = xndarray(inacc, axes_names=an, axes_domains=ad)

        return outputs

    def manageMapping(self, cuboid):
        pass

    def cleanObservables(self):
        del self.cumul3
        del self.cumul
        del self.currentValue

    def registerNbIterations(self, nbIt):
        pass

    def setSamplerEngine(self, sampler):
        # TODO : comment
        self.samplerEngine = sampler

    def linkToData(self,):
        raise NotImplementedError()

    def checkAndSetInitValue(self, variables):
        if self.currentValue is None:
            print 'Error - ', self.name, ' - initial value not set !'

    def samplingWarmUp(self, variables):
        """
        Called before the launch of the main sampling loop by the sampler
        engine. Should be overriden and perform precalculations.
        """
        pass

    def finalizeSampling(self):
        #from scikits.talkbox.tools.correlations import  acorr
        from pyhrf.stats import acorr
        if 0 and self.smplHistory is not None:
            logger.info('Compute autocorrelation of %s samples, shape=%s',
                        self.name, str(self.smplHistory.shape))
            trajectory = self.smplHistory[self.samplerEngine.nbSweeps::2]
            self.autocorrelation = acorr(trajectory)
            sn = trajectory.shape[0] ** .5
            t95 = 1.959963984540054 / sn
            self.autocorrelation_test = np.zeros(self.autocorrelation.shape,
                                                 dtype=np.int32)
            self.autocorrelation_test[np.where(self.autocorrelation > t95)] = 1
            self.autocorrelation_test[
                np.where(self.autocorrelation < -t95)] = -1

            self.autocorrelation_thresh = t95

            from scipy.stats import norm
            self.autocorrelation_pvalue = np.zeros(self.autocorrelation.shape,
                                                   dtype=np.float32)
            m_pos = np.where(self.autocorrelation > 0)
            if len(m_pos[0]) > 0:
                ac_pos = self.autocorrelation[m_pos]
                self.autocorrelation_pvalue[m_pos] = norm.sf(ac_pos * sn)

            m_neg = np.where(self.autocorrelation < 0)
            if len(m_neg[0]) > 0:
                ac_neg = self.autocorrelation[m_neg]
                self.autocorrelation_pvalue[m_neg] = norm.cdf(ac_neg * sn)

            logger.info('Compute posterior median for %s ', self.name)
            self.median = np.median(self.smplHistory, axis=0)

        self.check_final_value()

    def check_final_value(self):
        report = {}
        if self.samplerEngine.check_ftval is not None:
            if self.trueValue is None:
                logger.info('Warning: no true val to check against for %s',
                            self.name)
            elif self.sampleFlag:
                fv = self.get_final_value()
                tv = self.get_true_value()
                rtol = 0.1      # Relative tolerance value: 10%
                # Absolute tolerance value. TODO: dependency to parameter
                atol = 0.1

                # report['true_value'] = tv
                # report['final_value'] = fv
                abs_error = np.abs(tv - fv)
                report['abs_error'] = abs_error
                rel_error = abs_error / np.maximum(np.abs(tv), np.abs(fv))
                report['rel_error'] = rel_error

                report['accuracy'] = self.get_accuracy(abs_error, rel_error,
                                                       fv, tv, atol, rtol)
                is_accurate = report['accuracy'][1].all()
                logger.info('Fit error for %s: aerr=%f, rerr=%f, is_accurate=%s',
                            self.name, abs_error.mean(), rel_error.mean(),
                            is_accurate)
                if not is_accurate:
                    m = "Final value of %s is not close to " \
                        "true value.\n -> aerror: %s\n -> rerror: %s\n" \
                        " Final value:\n %s\n True value:\n %s\n" \
                        % (self.name, array_summary(report['abs_error']),
                           array_summary(report['rel_error']),
                           str(fv), str(tv))
                    if self.samplerEngine.check_ftval == 'raise':
                        raise Exception(m)
                    elif self.samplerEngine.check_ftval == 'print':
                        print '\n'.join(['!! ' + s for s in m.split('\n')])

        self.report_check_ft_val = report

    def get_accuracy(self, abs_error, rel_error, fv, tv, atol, rtol):
        """ Return the accuray of the estimate *fv*, compared to the true
        value *tv*

        Output:
            axes_names (list of str), accuracy (numpy array of booleans)
        """
        # same criterion as np.allclose:
        acc = abs_error <= (atol + rtol * np.maximum(np.abs(tv),
                                                     np.abs(fv)))
        return self.axes_names, acc

    def get_final_value(self):
        return self.finalValue

    def get_true_value(self):
        return self.trueValue

    def setFinalValue(self):
        self.finalValue = self.getMean()
        # if self.name=='nrl_by_session':
        # print 'self.cumul:', self.cumul[2,2,15]

    def sampleNextAlt(self, variables):
        """
        Define the behaviour of the variable at each sampling step when its
        sampling is not activated.
        """
        # print self.name, ' skip sampling ...'
        pass

    def sampleNextInternal(self, variables):
        """
        Define the behaviour of the variable at each sampling step when its
        sampling is not activated. Must be overriden in child classes.
        """
        raise NotImplementedError()

    def getMean(self):  # , itStart=None, itEnd=None, pace=1):
        """
        Wip ...
        Compute mean over MCMC iterations within the window defined by itStart,
        itEnd and pace. By default itStart is set to 'nbSweeps' and itEnd to
        the last iteration.
        """
        # print 'seeelf:', self, self.mean
        return self.mean
# if itStart == None :
##            itStart = self.samplerEngine.nbSweeps
# if itEnd == None :
##            itEnd = len(self.valHistory)-1
# print 'getMean '
# print 'itStart = ', itStart
# print 'itEnd = ', itEnd
# print 'range = ', arange(itStart, itEnd+1, pace)
# return self.valHistory[arange(itStart, itEnd+1, pace)].mean(0)

    def getMeanHistory(self):  # , itStart=None, histPace=1):
        if self.meanHistory is None:
            return (None, None)
        else:
            return (self.obsHistoryIts, self.meanHistory)

# if itStart == None :
##            itStart = self.samplerEngine.nbSweeps

##        itMax = len(self.valHistory)
##        meanHistory = [self.getMean(itStart, itStart+1)]
# for ite in xrange(itStart+1, itMax, histPace):
##            m = self.getMean(itStart, ite)
# if m != None:
##                meanHistory = concatenate( (meanHistory,[m]) )

##        itAxis = range(itStart, itMax, histPace)
# return (itAxis,meanHistory)

    # TODO idem as mean with variance ...


class GSDefaultCallbackHandler(xmlio.XmlInitable):
    """
    Class handling default action after Gibbs Sampling step (nothing). Should be inherited to define more specialized actions (such as plotting and reporting).

    """

    def __init__(self):
        xmlio.XmlInitable.__init__(self)

    def callback(self, it, vars, samplerEngine):
        """
        Execute action to be made after each Gibbs Sampling step (here : nothing). Should be overriden to define more specialized actions.
        @param it: the number of iterations elapsed in the current sampling process.
        @param samplerEngine: the parent gibbs sampler object
        @param vars: variables envolved in the sampling process (list of C{GibbsSamplerVariable} whose index is defined in L{samplerEngine})
        """
        pass

    def __call__(self, it, v, e):
        self.callback(it, v, e)


# class BayesFactorCallback(xmlio.XMLParamDrivenClass):

    # defaultParameters = {
        #'monParametre' : 458.,
        #}

    # def __init__(self, parameters=None, xmlHandler=xmlio.TypedXMLHandler(),
        # xmlLabel=None, xmlComment=None):
        # xmlio.XMLParamDrivenClass.__init__(self, parameters, xmlHandler,
        # xmlLabel, xmlComment)

        # Recuperation des parametres:
        #self.monParam = self.parameters['monParametre']
        # print 'BayesFactorCallback.__init__ :',
        # print 'self.monParam :', self.monParam

    # def callback(self, it, vars, samplerEngine):

        #samplerEngine.invLike = 1.
        #LVformer = 1.
        # N = samplerEngine.dataInput.varMBY.shape[0]       # N is the number of rows of P or H or yj
        #M = samplerEngine.dataInput.nbConditions
        # Q = samplerEngine.dataInput.colP       # Q is the number of columns of P
        #delta = samplerEngine.dataInput.delta
        #shrf = samplerEngine.get_variable('hrf')
        #snrl = samplerEngine.get_variable('nrl')
        #snoise = samplerEngine.get_variable('noise')
        # h = shrf.currentValue # shape = (nbCoeffHrf)
        # a = snrl.currentValue # shape = (nbConditions, nbVoxels)
        # y = samplerEngine.dataInput.varMBY # shape = (nbScans, nbVoxels)
        #H = shrf.varXh
        # noiseVars = snoise.currentValues # sh = (nbVoxels)
        #In = eye(N, dtype=float)

        # print 'Iteration :', it
        # if  it >= samplerEngine.nbSweeps:
        #shrf = samplerEngine.get_variable('hrf')

        # LVcurrent_value = LVformer*(0.5*(N-Q-M)-2)!*(0.5
        #samplerEngine.invLike = samplerEngine.invLike*(it-1)/it + LVcurrent_value/it
        #samplerEngine.Like = 1/samplerEngine.Like
        #LVformer = LVcurrent_value

        # print ' -> hrf value:', shrf.currentValue
        # print ' -> nrl value:', snrl.currentValue

        # print 'Harmonic mean estimate of the integrated likelihood...'

    # def finalize(self):
        # pass


class GSPrintCallbackHandler(GSDefaultCallbackHandler):
    """
    Class defining behaviour after each Gibbs Sampling step : printing reports to stdout.

    """

    def __init__(self, pace):
        self.pace = pace

    def callback(self, it, variables, samplerEngine):
        if not it % self.pace:
            print 'Iteration : ', it
            for v in variables:
                print v.name
                print v.currentValue
                print ' '
            print ' '



# -*- coding: utf-8 -*-

"""This module implements the VEM for BOLD data.

The function uses the C extension for expectation and maximization steps (see
src/pyhrf/vbjde/utilsmodule.c file).


See Also
--------
pyhrf.ui.analyser_ui, pyhrf.ui.treatment, pyhrf.ui.jde, pyhrf.ui.vb_jde_analyser

Notes
-----
TODO: add some refs?

Attributes
----------
eps : float
    mimics the machine epsilon to avoid zero values
logger : logger
    logger instance identifying this module to log informations

"""

import time
import logging

from collections import OrderedDict

import numpy as np

from scipy.spatial.distance import mahalanobis

import pyhrf.vbjde.vem_tools as vt

from pyhrf.boldsynth.hrf import getCanoHRF

logger = logging.getLogger(__name__)
eps = np.spacing(1)


def jde_vem_bold(graph, bold_data, onsets, durations, hrf_duration, nb_classes, tr, beta, dt, estimate_sigma_h=True,
                 sigma_h=0.05, it_max=-1, it_min=0, estimate_beta=True, contrasts=None, compute_contrasts=False,
                 hrf_hyperprior=0, estimate_hrf=True, constrained=False, zero_constraint=True, drifts_type="poly",
                 seed=6537546):
    """This is the main function that computes the VEM analysis on BOLD data.
    This function uses optimized python functions.

    Parameters
    ----------
    graph : ndarray of lists
        represents the neighbours indexes of each voxels index
    bold_data : ndarray, shape (nb_scans, nb_voxels)
        raw data
    onsets : dict
        dictionnary of onsets
    durations : # TODO
        # TODO
    hrf_duration : float
        hrf total time duration (in s)
    nb_classes : int
        the number of classes to classify the nrls. This parameter is provided
        for development purposes as most of the algorithm implies two classes
    tr : float
        time of repetition
    beta : float
        the initial value of beta
    dt : float
        hrf temporal precision
    estimate_sigma_h : bool, optional
        toggle estimation of sigma H
    sigma_h : float, optional
        initial or fixed value of sigma H
    it_max : int, optional
        maximal computed iteration number
    it_min : int, optional
        minimal computed iteration number
    estimate_beta : bool, optional
        toggle the estimation of beta
    contrasts : OrderedDict, optional
        dict of contrasts to compute
    compute_contrasts : bool, optional
        if True, compute the contrasts defined in contrasts
    hrf_hyperprior : float
        # TODO
    estimate_hrf : bool, optional
        if True, estimate the HRF for each parcel, if False use the canonical HRF
    constrained : bool, optional
        if True, add a constrains the l2 norm of the HRF to 1
    zero_constraint : bool, optional
        if True, add zeros to the beginning and the end of the estimated HRF.
    drifts_type : str, optional
        set the drifts basis type used. Can be "poly" for polynomial or "cos"
        for cosine
    seed : int, optional
        seed used by numpy to initialize random generator number

    Returns
    -------
    loop : int
        number of iterations before convergence
    nrls_mean : ndarray, shape (nb_voxels, nb_conditions)
        Neural response level mean value
    hrf_mean : ndarray, shape (hrf_len,)
        Hemodynamic response function mean value
    hrf_covar : ndarray, shape (hrf_len, hrf_len)
        Covariance matrix of the HRF
    labels_proba : ndarray, shape (nb_conditions, nb_classes, nb_voxels)
        probability of voxels being in one class
    noise_var : ndarray, shape (nb_voxels,)
        estimated noise variance
    nrls_class_mean : ndarray, shape (nb_conditions, nb_classes)
        estimated mean value of the gaussians of the classes
    nrls_class_var : ndarray, shape (nb_conditions, nb_classes)
        estimated variance of the gaussians of the classes
    beta : ndarray, shape (nb_conditions,)
        estimated beta
    drift_coeffs : ndarray, shape (# TODO)
        estimated coefficient of the drifts
    drift : ndarray, shape (# TODO)
        estimated drifts
    contrasts_mean : ndarray, shape (nb_voxels, len(contrasts))
        Contrasts computed from NRLs
    contrasts_var : ndarray, shape (nb_voxels, len(contrasts))
        Variance of the contrasts
    compute_time : list
        computation time of each iteration
    compute_time_mean : float
        computation mean time over iterations
    nrls_covar : ndarray, shape (nb_conditions, nb_conditions, nb_voxels)
        # TODO
    stimulus_induced_signal : ndarray, shape (nb_scans, nb_voxels)
        # TODO
    mahalanobis_zero : float
        Mahalanobis distance between estimated hrf_mean and the null vector
    mahalanobis_cano : float
        Mahalanobis distance between estimated hrf_mean and the canonical HRF
    mahalanobis_diff : float
        difference between mahalanobis_cano and mahalanobis_diff
    mahalanobis_prod : float
        product of mahalanobis_cano and mahalanobis_diff
    ppm_a_nrl : ndarray, shape (nb_voxels,)
        The posterior probability map using an alpha
    ppm_g_nrl : ndarray, shape (nb_voxels,)
        # TODO
    ppm_a_contrasts : ndarray, shape (nb_voxels,)
        # TODO
    ppm_g_contrasts : ndarray, shape (nb_voxels,)
        # TODO
    variation_coeff : float
        coefficient of variation of the HRF
    free_energy : list
        # TODO

    Notes
    -----
        See `A novel definition of the multivariate coefficient of variation
        <http://onlinelibrary.wiley.com/doi/10.1002/bimj.201000030/abstract>`_
        article for more information about the coefficient of variation.
    """

    logger.info("VEM started.")

    if not contrasts:
        contrasts = OrderedDict()

    np.random.seed(seed)

    nb_2_norm = 1
    normalizing = False
    regularizing = False

    if it_max <= 0:
        it_max = 100

    gamma = 7.5

    # Initialize sizes vectors
    hrf_len = np.int(np.ceil(hrf_duration / dt)) + 1

    nb_conditions = len(onsets)
    nb_scans = bold_data.shape[0]
    nb_voxels = bold_data.shape[1]
    X, occurence_matrix, condition_names = vt.create_conditions(onsets, durations, nb_conditions, nb_scans, hrf_len, tr,
                                                                dt)

    neighbours_indexes = vt.create_neighbours(graph)

    order = 2
    if regularizing:
        regularization = np.ones(hrf_len)
        regularization[hrf_len // 3:hrf_len // 2] = 2
        regularization[hrf_len // 2:2 * hrf_len // 3] = 5
        regularization[2 * hrf_len // 3:3 * hrf_len // 4] = 7
        regularization[3 * hrf_len // 4:] = 10
        # regularization[hrf_len//2:] = 10
    else:
        regularization = None

    d2 = vt.buildFiniteDiffMatrix(order, hrf_len, regularization)
    hrf_regu_prior_inv = d2.T.dot(d2) / pow(dt, 2 * order)

    if estimate_hrf and zero_constraint:
        hrf_len = hrf_len - 2
        hrf_regu_prior_inv = hrf_regu_prior_inv[1:-1, 1:-1]
        occurence_matrix = occurence_matrix[:, :, 1:-1]

    noise_struct = np.identity(nb_scans)

    noise_var = np.ones(nb_voxels)

    if nb_classes != 2:
        logger.warn('The number of classes is different to two.')

    labels_proba = np.zeros((nb_conditions, nb_classes, nb_voxels), dtype=np.float64)
    logger.info("Labels are initialized by setting everything to {}".format(1. / nb_classes))
    labels_proba[:, :, :] = 1. / nb_classes

    m_h = getCanoHRF(hrf_duration, dt)[1][:hrf_len]
    hrf_mean = np.array(m_h).astype(np.float64)

    if estimate_hrf:
        hrf_covar = np.identity(hrf_len, dtype=np.float64)
    else:
        hrf_covar = np.zeros((hrf_len, hrf_len), dtype=np.float64)

    beta = beta * np.ones(nb_conditions, dtype=np.float64)
    beta_list = [beta.copy()]

    if drifts_type == "poly":
        drift_basis = vt.poly_drifts_basis(nb_scans, 4, tr)
    elif drifts_type == "cos":
        drift_basis = vt.cosine_drifts_basis(nb_scans, 64, tr)
    else:
        raise Exception('drift type "%s" is not supported' % drifts_type)

    drift_coeffs = vt.drifts_coeffs_fit(bold_data, drift_basis)
    drift = drift_basis.dot(drift_coeffs)
    bold_data_drift = bold_data - drift

    # Parameters Gaussian mixtures
    nrls_class_mean = 2 * np.ones((nb_conditions, nb_classes))
    nrls_class_mean[:, 0] = 0

    nrls_class_var = 0.3 * np.ones((nb_conditions, nb_classes), dtype=np.float64)

    nrls_mean = (np.random.normal(nrls_class_mean, nrls_class_var)[:, :, np.newaxis] * labels_proba).sum(axis=1).T

    nrls_covar = np.identity(nb_conditions)[:, :, np.newaxis] + np.zeros((1, 1, nb_voxels))

    thresh_free_energy = 1e-4
    free_energy = [1.]
    free_energy_crit = [1.]

    compute_time = []
    start_time = time.time()
    loop = 0
    while (loop <= it_min
           or ((np.asarray(free_energy_crit[-5:]) > thresh_free_energy).any() and loop < it_max)):

        logger.info("{:-^80}".format(" Iteration n°" + str(loop + 1) + " "))

        logger.info("Expectation A step...")
        logger.debug("Before: nrls_mean = %s, nrls_covar = %s", nrls_mean, nrls_covar)
        nrls_mean, nrls_covar = vt.nrls_expectation(hrf_mean, nrls_mean, occurence_matrix, noise_struct, labels_proba,
                                                    nrls_class_mean, nrls_class_var, nb_conditions, bold_data_drift,
                                                    nrls_covar, hrf_covar, noise_var)
        logger.debug("After: nrls_mean = %s, nrls_covar = %s", nrls_mean, nrls_covar)

        logger.info("Expectation Z step...")
        logger.debug("Before: labels_proba = %s, labels_proba = %s", labels_proba, labels_proba)
        labels_proba = vt.labels_expectation(nrls_covar, nrls_mean, nrls_class_var, nrls_class_mean, beta, labels_proba,
                                             neighbours_indexes, nb_conditions, nb_classes, nb_voxels, parallel=True)
        logger.debug("After: labels_proba = %s, labels_proba = %s", labels_proba, labels_proba)

        if estimate_hrf:
            logger.info("Expectation H step...")
            logger.debug("Before: hrf_mean = %s, hrf_covar = %s", hrf_mean, hrf_covar)
            hrf_mean, hrf_covar = vt.hrf_expectation(nrls_covar, nrls_mean, occurence_matrix, noise_struct,
                                                     hrf_regu_prior_inv, sigma_h, nb_voxels, bold_data_drift, noise_var)

            if constrained:
                hrf_mean = vt.norm1_constraint(hrf_mean, hrf_covar)
                hrf_covar[:] = 0

            logger.debug("After: hrf_mean = %s, hrf_covar = %s", hrf_mean, hrf_covar)

            # Normalizing H at each nb_2_norm iterations:
            if not constrained and normalizing:
                # Normalizing is done before sigma_h, nrls_class_mean and nrls_class_var estimation
                # we should not include them in the normalisation step
                if (loop + 1) % nb_2_norm == 0:
                    hrf_norm = np.linalg.norm(hrf_mean)
                    hrf_mean /= hrf_norm
                    hrf_covar /= hrf_norm ** 2
                    nrls_mean *= hrf_norm
                    nrls_covar *= hrf_norm ** 2

        if estimate_hrf and estimate_sigma_h:
            logger.info("Maximization sigma_H step...")
            logger.debug("Before: sigma_h = %s", sigma_h)
            if hrf_hyperprior > 0:
                sigma_h = vt.maximization_sigmaH_prior(hrf_len, hrf_covar, hrf_regu_prior_inv, hrf_mean, hrf_hyperprior)
            else:
                sigma_h = vt.maximization_sigmaH(hrf_len, hrf_covar, hrf_regu_prior_inv, hrf_mean)
            logger.debug("After: sigma_h = %s", sigma_h)

        logger.info("Maximization (mu,sigma) step...")
        logger.debug("Before: nrls_class_mean = %s, nrls_class_var = %s", nrls_class_mean, nrls_class_var)
        nrls_class_mean, nrls_class_var = vt.maximization_class_proba(labels_proba, nrls_mean, nrls_covar)
        logger.debug("After: nrls_class_mean = %s, nrls_class_var = %s", nrls_class_mean, nrls_class_var)

        logger.info("Maximization L step...")
        logger.debug("Before: drift_coeffs = %s", drift_coeffs)
        drift_coeffs = vt.maximization_drift_coeffs(bold_data, nrls_mean, occurence_matrix, hrf_mean, noise_struct,
                                                    drift_basis)
        logger.debug("After: drift_coeffs = %s", drift_coeffs)

        drift = drift_basis.dot(drift_coeffs)
        bold_data_drift = bold_data - drift
        if estimate_beta:
            logger.info("Maximization beta step...")
            for cond_nb in xrange(0, nb_conditions):
                beta[cond_nb], success = vt.beta_maximization(beta[cond_nb] * np.ones((1,)),
                                                              labels_proba[cond_nb, :, :], neighbours_indexes, gamma)
            beta_list.append(beta.copy())
            logger.debug("beta = %s", str(beta))

        logger.info("Maximization sigma noise step...")
        noise_var = vt.maximization_noise_var(occurence_matrix, hrf_mean, hrf_covar, nrls_mean, nrls_covar,
                                              noise_struct, bold_data_drift, nb_scans)

        # Computing Free Energy
        free_energy.append(vt.free_energy_computation(nrls_mean, nrls_covar, hrf_mean, hrf_covar, hrf_len, labels_proba,
                                                      bold_data_drift, occurence_matrix, noise_var, noise_struct,
                                                      nb_conditions, nb_voxels, nb_scans, nb_classes, nrls_class_mean,
                                                      nrls_class_var, neighbours_indexes, beta, sigma_h,
                                                      np.linalg.inv(hrf_regu_prior_inv), hrf_regu_prior_inv, gamma,
                                                      hrf_hyperprior))

        free_energy_crit.append(abs((free_energy[-2] - free_energy[-1]) / free_energy[-2]))

        logger.info("Convergence criteria: %f (Threshold = %f)", free_energy_crit[-1], thresh_free_energy)
        loop += 1
        compute_time.append(time.time() - start_time)

    compute_time_mean = compute_time[-1] / loop

    mahalanobis_zero = np.nan
    mahalanobis_cano = np.nan
    mahalanobis_diff = np.nan
    mahalanobis_prod = np.nan
    variation_coeff = np.nan

    if estimate_hrf and not constrained and not normalizing:
        hrf_norm = np.linalg.norm(hrf_mean)
        hrf_mean /= hrf_norm
        hrf_covar /= hrf_norm ** 2
        sigma_h /= hrf_norm ** 2
        nrls_mean *= hrf_norm
        nrls_covar *= hrf_norm ** 2
        nrls_class_mean *= hrf_norm
        nrls_class_var *= hrf_norm ** 2
        mahalanobis_zero = mahalanobis(hrf_mean, np.zeros_like(hrf_mean), np.linalg.inv(hrf_covar))
        mahalanobis_cano = mahalanobis(hrf_mean, m_h, np.linalg.inv(hrf_covar))
        mahalanobis_diff = mahalanobis_cano - mahalanobis_zero
        mahalanobis_prod = mahalanobis_cano * mahalanobis_zero
        variation_coeff = np.sqrt((hrf_mean.T.dot(hrf_covar).dot(hrf_mean)) / (hrf_mean.T.dot(hrf_mean)) ** 2)

    if estimate_hrf and zero_constraint:
        hrf_mean = np.concatenate(([0], hrf_mean, [0]))

        # when using the zero constraint the hrf covariance is fill with
        # arbitrary zeros around the matrix, this is maybe a bad idea if we need
        # it for later computation...
        hrf_covar = np.concatenate((np.zeros((hrf_covar.shape[0], 1)), hrf_covar, np.zeros((hrf_covar.shape[0], 1))),
                                   axis=1)

        hrf_covar = np.concatenate((np.zeros((1, hrf_covar.shape[1])), hrf_covar, np.zeros((1, hrf_covar.shape[1]))),
                                   axis=0)

    if estimate_hrf:
        (delay_of_response, delay_of_undershoot, dispersion_of_response, dispersion_of_undershoot, ratio_resp_under,
         delay) = vt.fit_hrf_two_gammas(hrf_mean, dt, hrf_duration)
    else:
        (delay_of_response, delay_of_undershoot, dispersion_of_response, dispersion_of_undershoot, ratio_resp_under,
         delay) = (None, None, None, None, None, None)

    ppm_a_nrl, ppm_g_nrl = vt.ppms_computation(nrls_mean, np.diagonal(nrls_covar), nrls_class_mean, nrls_class_var,
                                               threshold_a="intersect")

    # Calculate contrast maps and variance
    nb_contrasts = len(contrasts)
    if compute_contrasts and nb_contrasts > 0:
        logger.info('Computing contrasts ...')
        (contrasts_mean, contrasts_var, contrasts_class_mean,
         contrasts_class_var) = vt.contrasts_mean_var_classes(contrasts, condition_names, nrls_mean, nrls_covar,
                                                              nrls_class_mean, nrls_class_var, nb_contrasts,
                                                              nb_classes, nb_voxels)

        ppm_a_contrasts, ppm_g_contrasts = vt.ppms_computation(contrasts_mean, contrasts_var, contrasts_class_mean,
                                                               contrasts_class_var)
        logger.info('Done computing contrasts.')
    else:
        (contrasts_mean, contrasts_var, contrasts_class_mean, contrasts_class_var, ppm_a_contrasts,
         ppm_g_contrasts) = (None, None, None, None, None, None)

    logger.info("Number of iterations to reach criterion: %d", loop)
    logger.info("Computational time = {t[0]:.0f} min {t[1]:.0f} s".format(t=divmod(compute_time[-1], 60)))
    logger.debug('nrls_class_mean: %s', nrls_class_mean)
    logger.debug('nrls_class_var: %s', nrls_class_var)
    logger.debug("sigma_H = %s", str(sigma_h))
    logger.debug("beta = %s", str(beta))

    stimulus_induced_signal = vt.computeFit(hrf_mean, nrls_mean, X, nb_voxels, nb_scans)
    snr = 20 * np.log(np.linalg.norm(bold_data.astype(np.float)) /
                      np.linalg.norm((bold_data_drift - stimulus_induced_signal).astype(np.float)))
    snr /= np.log(10.)
    logger.info('SNR comp = %f', snr)

    return (loop, nrls_mean, hrf_mean, hrf_covar, labels_proba, noise_var, nrls_class_mean, nrls_class_var, beta,
            drift_coeffs, drift, contrasts_mean, contrasts_var, compute_time[2:], compute_time_mean, nrls_covar,
            stimulus_induced_signal, mahalanobis_zero, mahalanobis_cano, mahalanobis_diff, mahalanobis_prod, ppm_a_nrl,
            ppm_g_nrl, ppm_a_contrasts, ppm_g_contrasts, variation_coeff, free_energy[1:], free_energy_crit[1:],
            beta_list[1:], delay_of_response, delay_of_undershoot, dispersion_of_response, dispersion_of_undershoot,
            ratio_resp_under, delay)



# -*- coding: utf-8 -*-

import os.path as op
import os
import unittest
import shutil
import difflib

import numpy as np
import numpy.testing as npt

try:
    os.environ["DISPLAY"]
except KeyError:
    import matplotlib
    matplotlib.use("Agg")

from pyhrf.ndarray import *
from pyhrf.tools import add_suffix

debug = False


class xndarrayTest(unittest.TestCase):

    def setUp(self):
        self.arr1d = np.array([5, 6, 10])
        self.arr1dNames = ['time']
        self.arr1dDom = {'time': np.array(['jan', 'feb', 'mar'])}
        self.arr1dLabel = 'growth'

        self.sh3d = (4, 5, 6)
        self.arr3d = np.arange(np.prod(self.sh3d)).reshape(self.sh3d)

        self.arr3dNames = ['axial', 'coronal', 'sagittal']
        self.arr3dDom = {'axial': np.arange(self.sh3d[0]) * 3,
                         'coronal': np.arange(self.sh3d[1]) * 3,
                         'sagittal': np.arange(self.sh3d[2]) * 3,
                         }
        self.arr3dLabel = 'intensity_value'

        self.mask2d = np.array([[0, 0, 0, 0, 0],
                                [0, 0, 0, 0, 0],
                                [0, 0, 1, 1, 0],
                                [0, 1, 1, 0, 0],
                                [0, 0, 1, 1, 0],
                                [0, 1, 1, 1, 0],
                                [0, 0, 0, 0, 0], ], dtype=int)

        self.mask2d_axes = ['x', 'y']

        nb_positions = self.mask2d.sum()
        conditions = np.array(['audio', 'video', 'computation', 'sentence'])
        nb_conditions = len(conditions)
        self.arr_flat_sh = (nb_positions, nb_conditions)

        sh = self.arr_flat_sh
        self.arr_flat = np.arange(np.prod(sh)).reshape(sh)
        self.arr_flat_names = ['position', 'condition']
        self.arr_flat_domains = {'condition': conditions}

        self.tmp_dir = pyhrf.get_tmp_path()

    def tearDown(self):
        shutil.rmtree(self.tmp_dir)

    def test_init(self):
        c = xndarray(self.arr1d, self.arr1dNames, self.arr1dDom,
                     self.arr1dLabel)
        if debug:
            print 'c:'
            print c.descrip()

    def test_squeeze(self):

        c = xndarray(np.arange(10).reshape(1, 5, 1, 2, 1),
                     ['a', 'b', 'c', 'd', 'e'], {'a': [2], 'b': np.arange(5) * 3, 'e': [5]})

        cs = c.squeeze()

        self.assertEqual(cs.axes_names, ['b', 'd'])

        cs2 = c.squeeze_all_but(['a', 'b', 'c'])

        self.assertEqual(cs2.axes_names, ['a', 'b', 'c', 'd'])

    def test_to_latex_1d(self):
        c = xndarray(self.arr1d, self.arr1dNames, self.arr1dDom,
                     self.arr1dLabel)

        if debug:
            print 'c:'
            print c.descrip()
            print self.arr1d

        result = c.to_latex(col_axes=['time'], val_fmt='%d')

        good_result = \
            '\\begin{tabular}{c c c}\n' \
            '\multicolumn{3}{c}{time}\\\\\n' \
            'jan & feb & mar\\\\\n' \
            '5 & 6 & 10\\\\\n' \
            '\\end{tabular}'

        self.assertEqual(result, good_result,
                         'to latex returned:\n' + result + '\n' + 'expected:\n' + good_result)

    def test_to_latex_3d(self):

        sh = (2, 3, 4)
        c = xndarray(np.arange(np.prod(sh)).reshape(sh), ['mod', 'time', 'subj'],
                     {'mod': ['m1', 'm2'],
                      'time': np.arange(3) * .5,
                      'subj': ['s1', 's2', 's3', 's4']})

        if debug:
            print 'c:'
            print c.descrip()

        result = c.to_latex(col_axes=['mod', 'time', 'subj'],
                            val_fmt='%d')

        good_result = \
            '\\begin{tabular}{' + ' '.join(['c'] * 24) + '}\n' \
            '\multicolumn{24}{c}{mod}\\\\\n' \
            '\multicolumn{12}{c}{m1} & \multicolumn{12}{c}{m2}\\\\\n' \
            '\multicolumn{24}{c}{time}\\\\\n' \
            '\multicolumn{4}{c}{0.0} & \multicolumn{4}{c}{0.5} & '\
            '\multicolumn{4}{c}{1.0} & \multicolumn{4}{c}{0.0} & '\
            '\multicolumn{4}{c}{0.5} & \multicolumn{4}{c}{1.0}\\\\\n' \
            '\multicolumn{24}{c}{subj}\\\\\n' \
            's1 & s2 & s3 & s4 & s1 & s2 & s3 & s4 & s1 & s2 & s3 & s4 '\
            '& s1 & s2 & s3 & s4 & s1 & s2 & s3 & s4 & s1 & s2 & s3 & s4\\\\\n' + \
            ' & '.join([str(i) for i in np.arange(np.prod(sh), dtype=int)]) + \
            '\\\\\n' + \
            '\\end{tabular}'

        self.assertEqual(result, good_result,
                         'to latex returned:\n' + result + '\n' + 'expected:\n' + good_result)

        result = c.to_latex(col_axes=['mod', 'subj'], row_axes=['time'],
                            hval_fmt={'time': '%1.1f'}, val_fmt='%d')

        l1 = range(4) + range(12, 16)
        l2 = range(4, 8) + range(16, 20)
        l3 = range(8, 12) + range(20, 24)
        good_result = \
            '\\begin{tabular}{c c | ' + ' '.join(['c'] * 8) + '}\n' \
            '&&\multicolumn{8}{c}{mod}\\\\\n' \
            '&&\multicolumn{4}{c}{m1} & \multicolumn{4}{c}{m2}\\\\\n' \
            '&&\multicolumn{8}{c}{subj}\\\\\n' \
            '&&s1 & s2 & s3 & s4 & s1 & s2 & s3 & s4\\\\\n' + \
            '\multirow{3}{*}{time} & 0.0 & ' + \
            ' & '.join([str(i) for i in l1]) + '\\\\\n' + \
            ' & 0.5 & ' + ' & '.join([str(i) for i in l2]) + '\\\\\n' + \
            ' & 1.0 & ' + ' & '.join([str(i) for i in l3]) + '\\\\\n' + \
            '\\end{tabular}'

        self.assertEqual(result, good_result,
                         'to latex returned:\n' + result + '\n' + 'expected:\n' + good_result)

        result = c.to_latex(col_axes=['subj'], row_axes=['mod', 'time'],
                            hval_fmt={'time': '%1.1f'}, val_fmt='%d')

        lines = [range(i, i + 4) for i in range(0, 24, 4)]
        good_result = \
            '\\begin{tabular}{c c c c | ' + ' '.join(['c'] * 4) + '}\n' \
            '&&&&\multicolumn{4}{c}{subj}\\\\\n' \
            '&&&&s1 & s2 & s3 & s4\\\\\n' + \
            '\multirow{6}{*}{mod} & \multirow{3}{*}{m1} & '\
            '\multirow{6}{*}{time} & 0.0 & ' + \
            ' & '.join([str(i) for i in lines[0]]) + '\\\\\n' + \
            ' &  &  & 0.5 & ' + ' & '.join([str(i) for i in lines[1]]) + '\\\\\n' + \
            ' &  &  & 1.0 & ' + ' & '.join([str(i) for i in lines[2]]) + '\\\\\n' + \
            ' & \multirow{3}{*}{m2} &  & 0.0 & ' + \
            ' & '.join([str(i) for i in lines[3]]) + '\\\\\n' + \
            ' &  &  & 0.5 & ' + ' & '.join([str(i) for i in lines[4]]) + '\\\\\n' + \
            ' &  &  & 1.0 & ' + ' & '.join([str(i) for i in lines[5]]) + '\\\\\n' + \
            '\\end{tabular}'

        self.assertEqual(result, good_result,
                         'to latex returned:\n' + result + '\n' + 'expected:\n' + good_result)

    def test_to_latex_3d_join_style(self):

        sh = (2, 3, 4)
        c = xndarray(np.arange(np.prod(sh)).reshape(sh), ['mod', 'time', 'subj'],
                     {'mod': ['m1', 'm2'],
                      'time': np.arange(3) * .5,
                      'subj': ['s1', 's2', 's3', 's4']})

        if debug:
            print 'c:'
            print c.descrip()

        result = c.to_latex(col_axes=['mod', 'subj'], row_axes=['time'],
                            header_styles={'mod': 'join', 'time': 'join'},
                            val_fmt='%d',
                            hval_fmt={'time': '%1.1f'})

        l1 = range(4) + range(12, 16)
        l2 = range(4, 8) + range(16, 20)
        l3 = range(8, 12) + range(20, 24)
        good_result = \
            '\\begin{tabular}{c | ' + ' '.join(['c'] * 8) + '}\n' \
            '&\multicolumn{4}{c}{mod=m1} & \multicolumn{4}{c}{mod=m2}\\\\\n' \
            '&\multicolumn{8}{c}{subj}\\\\\n' \
            '&s1 & s2 & s3 & s4 & s1 & s2 & s3 & s4\\\\\n' + \
            'time=0.0 & ' + \
            ' & '.join([str(i) for i in l1]) + '\\\\\n' + \
            'time=0.5 & ' + ' & '.join([str(i) for i in l2]) + '\\\\\n' + \
            'time=1.0 & ' + ' & '.join([str(i) for i in l3]) + '\\\\\n' + \
            '\\end{tabular}'

        self.assertEqual(result, good_result,
                         'to latex returned:\n' + result + '\n' + 'expected:\n' + good_result)

    def test_to_latex_3d_inner_axes(self):

        sh = (2, 3, 4)
        c = xndarray(np.arange(np.prod(sh)).reshape(sh), ['mod', 'time', 'subj'],
                     {'mod': ['m1', 'm2'],
                      'time': np.arange(3) * .5,
                      'subj': ['s1', 's2', 's3', 's4']})

        if debug:
            print 'c:'
            print c.descrip()

        result = c.to_latex(col_axes=['subj'], row_axes=['time'],
                            inner_axes=['mod'],
                            header_styles={'mod': 'join', 'time': 'join'},
                            val_fmt="%d",
                            hval_fmt={'time': "%1.1f"})

        l1 = ['%d | %d' % (a, b) for a, b in zip(range(4), range(12, 16))]
        l2 = ['%d | %d' % (a, b) for a, b in zip(range(4, 8), range(16, 20))]
        l3 = ['%d | %d' % (a, b) for a, b in zip(range(8, 12), range(20, 24))]
        good_result = \
            '\\begin{tabular}{c | ' + ' '.join(['c'] * 4) + '}\n' \
            '&\multicolumn{4}{c}{subj}\\\\\n' \
            '&s1 & s2 & s3 & s4\\\\\n' + \
            'time=0.0 & ' + \
            ' & '.join([str(i) for i in l1]) + '\\\\\n' + \
            'time=0.5 & ' + ' & '.join([str(i) for i in l2]) + '\\\\\n' + \
            'time=1.0 & ' + ' & '.join([str(i) for i in l3]) + '\\\\\n' + \
            '\\end{tabular}'

        self.assertEqual(result, good_result,
                         'to latex returned:\n' + result + '\n' + 'expected:\n' + good_result)

    def test_to_latex_3d_col_align(self):
        sh = (2, 3, 4)
        c = xndarray(np.arange(np.prod(sh)).reshape(sh), ['mod', 'time', 'subj'],
                     {'mod': ['m1', 'm2'],
                      'time': np.arange(3) * .5,
                      'subj': ['s1', 's2', 's3', 's4']})

        if debug:
            print 'c:'
            print c.descrip()

        result = c.to_latex(col_axes=['mod', 'subj'], row_axes=['time'],
                            header_styles={'mod': 'join', 'time': 'join'},
                            val_fmt='%d', col_align={'mod': 'c@{}'},
                            hval_fmt={'time': '%1.1f'})

        l1 = range(4) + range(12, 16)
        l2 = range(4, 8) + range(16, 20)
        l3 = range(8, 12) + range(20, 24)
        good_result = \
            '\\begin{tabular}{c | c c c c@{} c c c c@{}}\n' \
            '&\multicolumn{4}{c}{mod=m1} & \multicolumn{4}{c}{mod=m2}\\\\\n' \
            '&\multicolumn{8}{c}{subj}\\\\\n' \
            '&s1 & s2 & s3 & s4 & s1 & s2 & s3 & s4\\\\\n' + \
            'time=0.0 & ' + \
            ' & '.join([str(i) for i in l1]) + '\\\\\n' + \
            'time=0.5 & ' + ' & '.join([str(i) for i in l2]) + '\\\\\n' + \
            'time=1.0 & ' + ' & '.join([str(i) for i in l3]) + '\\\\\n' + \
            '\\end{tabular}'

        self.assertEqual(result, good_result,
                         'to latex returned:\n' + result + '\n' + 'expected:\n' + good_result)

    def test_to_latex_3d_hide_name_style(self):

        sh = (2, 3, 4)
        c = xndarray(np.arange(np.prod(sh)).reshape(sh), ['mod', 'time', 'subj'],
                     {'mod': ['m1', 'm2'],
                      'time': np.arange(3) * .5,
                      'subj': ['s1', 's2', 's3', 's4']})

        if debug:
            print 'c:'
            print c.descrip()

        result = c.to_latex(col_axes=['mod', 'subj'], row_axes=['time'],
                            header_styles={'mod': 'join', 'time': 'hide_name'},
                            hval_fmt={'time': '%1.1f'}, val_fmt='%d')

        l1 = range(4) + range(12, 16)
        l2 = range(4, 8) + range(16, 20)
        l3 = range(8, 12) + range(20, 24)
        good_result = \
            '\\begin{tabular}{c | ' + ' '.join(['c'] * 8) + '}\n' \
            '&\multicolumn{4}{c}{mod=m1} & \multicolumn{4}{c}{mod=m2}\\\\\n' \
            '&\multicolumn{8}{c}{subj}\\\\\n' \
            '&s1 & s2 & s3 & s4 & s1 & s2 & s3 & s4\\\\\n' + \
            '0.0 & ' + \
            ' & '.join([str(i) for i in l1]) + '\\\\\n' + \
            '0.5 & ' + ' & '.join([str(i) for i in l2]) + '\\\\\n' + \
            '1.0 & ' + ' & '.join([str(i) for i in l3]) + '\\\\\n' + \
            '\\end{tabular}'

        self.assertEqual(result, good_result,
                         'to latex returned:\n' + result + '\n' + 'expected:\n' + good_result)

    def test_combine_domains(self):
        c = xndarray(np.zeros((2, 3, 4)), ['mod', 'time', 'subj'],
                     {'mod': ['m1', 'm2'],
                      'time': np.arange(3) * .5,
                      'subj': ['s1', 's2', 's3', 's4']})

        cdom = c._combine_domains(['subj', 'time', 'mod'])
        expectation = [['s1', 's2', 's3', 's4'],
                       list(np.arange(3) * .5) * 4,
                       ['m1', 'm2'] * 4 * 3
                       ]
        self.assertEqual(cdom, expectation)

        cdom = c._combine_domains(['subj', 'time'])
        expectation = [['s1', 's2', 's3', 's4'],
                       list(np.arange(3) * .5) * 4,
                       ]
        self.assertEqual(cdom, expectation)

    def test_fill(self):
        """
        TODO
        """

        data = xndarray(np.array([[[1, 1, 2, 2],
                                   [2, 3, 3, 4],
                                   [4, 4, 4, 4]],
                                  [[2, 2, 3, 3],
                                   [3, 4, 4, 5],
                                   [5, 5, 5, 5]]]),
                        axes_names=['time', 'axial', 'sagital'])

        fill_data = xndarray(np.array([[10, 10, 10, 10],
                                       [20, 20, 20, 20],
                                       [30, 30, 30, 30]]),
                             axes_names=['axial', 'sagital'])
        data.fill(fill_data)

        npt.assert_array_equal(data.data[0], fill_data.data)
        npt.assert_array_equal(data.data[1], fill_data.data)

    def test_explode(self):
        mask = xndarray(np.array([[0, 0, 1, 1],
                                  [1, 7, 7, 3],
                                  [3, 3, 3, 3]]), axes_names=['axial', 'sagital'])

        data = xndarray(np.array([[[1, 1, 2, 2],
                                   [2, 3, 3, 4],
                                   [4, 4, 4, 4]],
                                  [[2, 2, 3, 3],
                                   [3, 4, 4, 5],
                                   [5, 5, 5, 5]]]),
                        axes_names=['time', 'axial', 'sagital'])

        exploded_data = data.explode(mask, new_axis='position')

        self.assertEqual(len(exploded_data), len(np.unique(mask.data)))
        self.assertEqual(exploded_data[0].axes_names, ['time', 'position'])
        npt.assert_array_equal(
            exploded_data[0].data, np.array([[1, 1], [2, 2]]))
        npt.assert_array_equal(
            exploded_data[7].data, np.array([[3, 3], [4, 4]]))
        npt.assert_array_equal(exploded_data[3].data, np.array([[4, 4, 4, 4, 4],
                                                                [5, 5, 5, 5, 5]]))

    def test_merge(self):
        """
        TODO !!!
        """
        mask = xndarray(np.array([[0, 0, 1, 1],
                                  [1, 7, 7, 3],
                                  [3, 3, 3, 3]]), axes_names=['axial', 'sagital'])

        data = xndarray(np.array([[[1, 1, 2, 2],
                                   [2, 3, 3, 4],
                                   [4, 4, 4, 4]],
                                  [[2, 2, 3, 3],
                                   [3, 4, 4, 5],
                                   [5, 5, 5, 5]]]),
                        axes_names=['time', 'axial', 'sagital'])

        exploded_data = data.explode(mask, new_axis='position')

        new_data = merge(exploded_data, mask, axis='position')

        npt.assert_array_equal(new_data.data, data.data)
        npt.assert_array_equal(new_data.axes_names, data.axes_names)

    def test_set_orientation(self):
        c = xndarray(
            self.arr3d, self.arr3dNames, self.arr3dDom, self.arr3dLabel)
        if debug:
            print 'Original cuboid:'
            print c.descrip()

        c.set_orientation(['sagittal', 'coronal', 'axial'])
        if debug:
            print 'After changing orientation to ', \
                ['sagittal', 'coronal', 'axial']
            print c.descrip()
        assert c.data.shape == (self.sh3d[2], self.sh3d[1], self.sh3d[0])

        c.set_orientation(['sagittal', 'axial', 'coronal'])
        if debug:
            print 'After changing orientation to ', \
                ['sagittal', 'axial', 'coronal']
            print c.descrip()
        assert c.data.shape == (self.sh3d[2], self.sh3d[0], self.sh3d[1])

    def test_xmapping(self):
        a = xndarray(np.arange(2 * 4).reshape(2, 4).T, ['time', 'parcel'],
                     {'time': np.arange(4) * .5, 'parcel': [2, 6]})
        parcel_map = xndarray(np.array([[2, 2, 2, 6], [6, 6, 6, 0], [6, 6, 0, 0]]),
                              ['axial', 'coronal'], value_label='parcel')
        a_mapped = a.map_onto(parcel_map)

        self.assertEqual(a_mapped.data.shape,
                         parcel_map.data.shape + a.data.shape[:1])
        self.assertEqual(a_mapped.axes_names, ['axial', 'coronal', 'time'])
        npt.assert_array_equal(a_mapped.get_domain('time'),
                               a.get_domain('time'))
        npt.assert_array_equal(a_mapped.data[0, 0], a.data[:, 0])
        npt.assert_array_equal(a_mapped.data[1, 0], a.data[:, 1])
        npt.assert_array_equal(a_mapped.data[-1, -1], 0.)

    def test_xmapping_inconsistent_mapping_value(self):
        a = xndarray(np.arange(2 * 4).reshape(2, 4).T, ['time', 'parcel'],
                     {'time': np.arange(4) * .5, 'parcel': [2, 6]})
        parcel_map = xndarray(np.array([[2, 2, 2, 6], [6, 6, 6, 0], [6, 6, 0, 0]]),
                              ['axial', 'coronal'], value_label='ROI')
        self.assertRaisesRegexp(ArrayMappingError,
                                r'Value label "ROI" of xmapping not found '
                                'in array axes \(time, parcel\)',
                                a.map_onto, parcel_map)

    def test_xmapping_inconsistent_domain(self):
        a = xndarray(np.arange(2 * 4).reshape(2, 4).T, ['time', 'parcel'],
                     {'time': np.arange(4) * .5, 'parcel': [2, 7]})
        parcel_map = xndarray(np.array([[2, 2, 2, 6], [6, 6, 6, 0], [6, 6, 0, 0]]),
                              ['axial', 'coronal'], value_label='parcel')
        self.assertRaisesRegexp(ArrayMappingError,
                                'Domain of axis "parcel" to be mapped is '
                                'not a subset of values in the mapping array.',
                                a.map_onto, parcel_map)

    def test_expansion(self):

        c = xndarray(self.arr_flat, self.arr_flat_names, self.arr_flat_domains)

        if debug:
            print 'Original cuboid:'
            print c.descrip()

        expanded_c = c.expand(self.mask2d, axis='position',
                              target_axes=self.mask2d_axes)
        if debug:
            print 'Expanded cuboid:'
            print expanded_c.descrip()

        assert expanded_c.data.shape == self.mask2d.shape + \
            self.arr_flat.shape[1:]

    def test_flatten_and_expand(self):

        sh = (2, 100, 50)
        c = xndarray(np.arange(np.prod(sh)).reshape(sh), ['c', 'voxel', 't'])
        m = np.zeros((10, 10, 10), dtype=int)
        m.flat[:100] = 1
        c_expanded = c.expand(m, axis='voxel', target_axes=['x', 'y', 'z'])

        c_flat = c_expanded.flatten(m, ['x', 'y', 'z'], 'voxel')
        assert c == c_flat

    def test_sub_cuboid(self):

        c = xndarray(self.arr_flat, self.arr_flat_names, self.arr_flat_domains)

        expanded_c = c.expand(self.mask2d, axis='position',
                              target_axes=self.mask2d_axes)
        if debug:
            print 'Original cuboid:'
            print expanded_c.descrip()

        sub_c = expanded_c.sub_cuboid(
            condition='audio', orientation=['y', 'x'])
        if debug:
            print 'Sub cuboid:'
            print sub_c.descrip()

        assert sub_c.data.shape == (self.mask2d.shape[1], self.mask2d.shape[0])

    def test_operations(self):

        data1 = np.arange(4 * 5, dtype=float).reshape(4, 5) + 1
        data1_copy = data1.copy()
        data2 = np.arange(4 * 5).reshape(4, 5) * 4. + 1
        data2_copy = data2.copy()

        c = xndarray(data1, ['x', 'time'], {'time': np.arange(5) + 2 / 3.})
        c_bak = c

        c2 = xndarray(data2, ['x', 'time'], {'time': np.arange(5) + 2 / 3.})

        c_add = c + c2
        assert (c_add.data == data1_copy + data2_copy).all()

        c += c2
        assert c.data is data1
        assert (c.data == data1_copy + data2_copy).all()
        assert c is c_bak

        c_sub = c - c2
        assert (c_sub.data == data1 - data2).all()

        c -= c2
        assert (c_sub.data == data1_copy).all()
        assert c is c_bak

        c_mul = c * c2
        assert (c_mul.data == data1_copy * data2_copy).all()

        c_div = c / c2
        assert (c_div.data.flat[1:] == (
            data1_copy / data2_copy).flat[1:]).all()

        r = 2 / c
        npt.assert_array_equal(r.data, 2 / c.data)

        r = 4 * c
        npt.assert_array_equal(r.data, 4 * c.data)

        r = 4 + c
        npt.assert_array_equal(r.data, 4 + c.data)

        r = 5 - c
        npt.assert_array_equal(r.data, 5 - c.data)

    def test_sub_cuboid_with_float_domain(self):

        c = xndarray(np.arange(4 * 2).reshape(4, 2), ['x', 'time'],
                     {'time': np.array([2 / 3., 2.3])})
        if debug:
            print 'Original cuboid:'
            print c.descrip()

        sub_c = c.sub_cuboid(time=2 / 3.)
        if debug:
            print 'sub_cuboid at time=2/3.:'
            print sub_c.descrip()

    def test_equality(self):
        c1 = xndarray(self.arr3d, self.arr3dNames, self.arr3dDom,
                      self.arr3dLabel)

        c2 = xndarray(self.arr3d, self.arr3dNames, self.arr3dDom,
                      self.arr3dLabel)

        assert c1 == c2

        c3 = xndarray(self.arr3d, self.arr3dNames)

        assert c1 != c3

    def test_save_as_nii(self):

        c = xndarray(self.arr3d, ['x', 'y', 'z'],
                     {'x': np.arange(self.sh3d[0]) * 3,
                      'y': np.arange(self.sh3d[1]) * 3,
                      'z': np.arange(self.sh3d[2]) * 3,
                      },
                     self.arr3dLabel)

        if debug:
            print 'Original cuboid:'
            print c.descrip()

        fn = op.join(self.tmp_dir, 'cuboid.nii')
        if debug:
            print 'fn:', fn

        c.save(fn)
        assert op.exists(fn)

        c_loaded = xndarray.load(fn)

        if debug:
            print 'Loaded cuboid:'
            print c_loaded.descrip()

        assert c == c_loaded

    def test_save_as_gii(self):

        c = xndarray(np.arange(4 * 2).reshape(4, 2), ['x', 'time'],
                     {'time': np.array([2 / 3., 2.3])})

        if debug:
            print 'Original cuboid:'
            print c.descrip()
            print c.data

        fn = op.join(self.tmp_dir, 'cuboid.gii')
        if debug:
            print 'fn:', fn

        c.save(fn)
        assert op.exists(fn)

        c_loaded = xndarray.load(fn)

        if debug:
            print 'Loaded cuboid:'
            print c_loaded.descrip()
            print c_loaded.data

        assert c == c_loaded

    def test_split(self):

        sh = (2, 4, 4, 4)
        c = xndarray(np.arange(np.prod(sh)).reshape(sh), ['condition'] + MRI3Daxes,
                     {'condition': ['audio', 'video']})
        if debug:
            print 'Original cub:'
            print c.descrip()

        fn = op.join(self.tmp_dir, 'cub.nii')
        if debug:
            print 'Save and load original cub'
        c.save(fn)
        c = xndarray.load(fn)

        fn = op.join(self.tmp_dir, 'cub2.nii')
        if debug:
            print 'Save and load new cub with meta data from original cuboid'
        sh = (4, 4, 4)
        c2 = xndarray(np.arange(np.prod(sh)).reshape(sh), MRI3Daxes,
                      meta_data=c.meta_data)
        c2.save(fn)
        c2 = xndarray.load(fn)

        fns = []
        sub_cuboids = []
        if debug:
            print 'Split and save sub cuboids'
        for dvalue, sub_c in c.split('condition').iteritems():
            fn = op.join(self.tmp_dir, add_suffix('sub_c.nii',
                                                  '_%s' % str(dvalue)))
            if debug and dvalue == 'audio':

                print 'fn_out:', fn
                print 'sub_c:'
                print sub_c.descrip()
                sub_cuboids.append(sub_c)
                sub_c.save(fn)
                fns.append(fn)
        if debug:
            print ''
            print 'Load sub c again ...'
        for fn, sub_c in zip(fns, sub_cuboids):
            if debug:
                print 'fn:', fn
            c_loaded = xndarray.load(fn)
            if debug:
                print 'c_loaded:'
                print c_loaded.descrip()
            self.assertEqual(c_loaded, sub_c)

    def test_stack(self):

        d1 = np.arange(4 * 5).reshape(4, 5)
        c1 = xndarray(d1, ['x', 'y'],
                      {'x': np.arange(4) * 3,
                       'y': np.arange(5) * 3,
                       })
        if debug:
            print 'c1:'
            print c1.descrip()

        d2 = np.arange(4 * 5).reshape(4, 5) * 2
        c2 = xndarray(d2, ['x', 'y'],
                      {'x': np.arange(4) * 3,
                       'y': np.arange(5) * 3,
                       })
        if debug:
            print 'c2:'
            print c2.descrip()

        c_stacked = stack_cuboids([c1, c2], 'stack_axis', ['c1', 'c2'])

        if debug:
            print 'c_stacked:'
            print c_stacked.descrip()

        a_stacked = np.array([d1, d2])
        assert (c_stacked.data == a_stacked).all()

    def test_tree_to_xndarray(self):
        from pyhrf.ndarray import xndarray, tree_to_xndarray
        from pyhrf.tools import set_leaf
        d1 = {}
        set_leaf(d1, ['1', '2.1', '3.1'], xndarray(np.array([1])))
        set_leaf(d1, ['1', '2.1', '3.2'], xndarray(np.array([2])))
        set_leaf(d1, ['1', '2.2', '3.1'], xndarray(np.array([3])))
        set_leaf(d1, ['1', '2.2', '3.2'], xndarray(np.array([3.1])))
        d2 = {}
        set_leaf(d2, ['1', '2.1', '3.1'], xndarray(np.array([10])))
        set_leaf(d2, ['1', '2.1', '3.2'], xndarray(np.array([11])))
        set_leaf(d2, ['1', '2.2', '3.1'], xndarray(np.array([12])))
        set_leaf(d2, ['1', '2.2', '3.2'], xndarray(np.array([13])))

        d = {'d1': d1, 'd2': d2}
        labels = ['case', 'p1', 'p2', 'p3']

        c = tree_to_xndarray(d, labels)
        self.assertEqual(c.data.shape, (2, 1, 2, 2, 1))

        npt.assert_array_equal(c.axes_domains['case'], ['d1', 'd2'])
        npt.assert_array_equal(c.axes_domains['p1'], ['1'])
        npt.assert_array_equal(c.axes_domains['p2'], ['2.1', '2.2'])
        npt.assert_array_equal(c.axes_domains['p3'], ['3.1', '3.2'])

    def test_cartesian_eval(self):
        """
        Test the multiple evaluations of a function that returns
        a xndarray, over the cartesian products of given arguments.
        """

        def foo(a, b, size, aname):
            return xndarray(np.ones(size) * a + b, axes_names=[aname])

        from pyhrf.tools import cartesian_apply
        from pyhrf.tools.backports import OrderedDict

        varying_args = OrderedDict([('a', [1, 2]), ('b', [.5, 1.5])])
        fixed_args = {'size': 2, 'aname': 'my_axis'}

        res = tree_to_xndarray(cartesian_apply(varying_args, foo, fixed_args))

        self.assertEqual(res.data.shape, (2, 2, 2))
        npt.assert_array_equal(res.data, np.array([[[1.5, 1.5],
                                                    [2.5, 2.5]],
                                                   [[2.5, 2.5],
                                                    [3.5, 3.5]]]))

    def test_unstack_2D(self):
        c = xndarray(np.arange(6).reshape(2, 3), axes_names=['a1', 'ia'],
                     axes_domains={'a1': ['out_dv1', 'out_dv2'],
                                   'ia': ['in_dv1', 'in_dv2', 'in_dv3']})
        uc = c.unstack(['a1'], ['ia'])

        self.assertEqual(uc.data.shape, (2,))
        self.assertEqual(uc.data[0].data.shape, (3,))
        self.assertEqual(len(uc.axes_domains), 1)
        npt.assert_array_equal(uc.axes_domains['a1'], ['out_dv1', 'out_dv2'])

    def test_unstack_empty_inner_axes(self):
        size = 6
        ad = ['out_dv%d' % i for i in range(size)]
        c = xndarray(
            np.arange(size), axes_names=['a1'], axes_domains={'a1': ad})
        uc = c.unstack(['a1'], [])

        self.assertEqual(uc.data.shape, (size,))
        self.assertEqual(len(uc.axes_domains), 1)
        npt.assert_array_equal(uc.axes_domains['a1'], ad)


class TestHtml(unittest.TestCase):

    def setUp(self):
        self.tmp_dir = pyhrf.get_tmp_path()

    def tearDown(self):
        shutil.rmtree(self.tmp_dir)

    def test_txt_1d_col_axes_only(self):
        a = xndarray([1, 2], ['measure'], {'measure': ['mon', 'tue']})
        html = a.to_html_table([], ['measure'], [])
        self.assertIsInstance(html, str)
        expected = '<table><tr><th colspan="2">measure</th></tr>' \
                   '<tr><th colspan="1">mon</th>' \
                   '<th colspan="1">tue</th></tr>' \
                   '<tr><td>1</td><td>2</td></tr></table>'
        self.assertEqual(html, expected)

    def test_txt_1d_row_axes_only(self):
        a = xndarray([1, 2], ['measure'], {'measure': ['mon', 'tue']})
        html = a.to_html_table(['measure'], [], [])
        self.assertIsInstance(html, str)
        expected = '<table><tr><th rowspan="2">' \
            '<div class="rotate">measure</div></th>' \
            '<th rowspan="1">mon</th><td>1</td></tr>' \
            '<tr><th rowspan="1">tue</th><td>2</td></tr>' \
            '</table>'
        self.assert_html_equal(html, expected)

    def test_txt_tooltip(self):
        a = xndarray([1, 2], ['measure'], {'measure': ['mon', 'tue']})
        html = a.to_html_table(['measure'], [], [], tooltip=True)
        self.assertIsInstance(html, str)
        expected = '<table><tr>'\
                   '<th rowspan="2">'\
                   '<div class="rotate">measure</div></th>' \
                   '<th rowspan="1">mon</th>'\
                   '<td title="measure=mon">1</td></tr>' \
                   '<tr><th rowspan="1">tue</th>'\
                   '<td title="measure=tue">2</td></tr>' \
                   '</table>'
        self.assert_html_equal(html, expected)

    def test_plot(self):
        sh = (2, 3, 4)
        a = xndarray(np.arange(np.prod(sh)).reshape(sh),
                     ['day', 'strength', 'position'],
                     {'day': ['mon', 'tue'],
                      'strength': [0., .5, 1.2],
                      'position': [0, 10, 20, 30]})

        fig_dir = op.join(self.tmp_dir, 'figs')
        os.makedirs(fig_dir)
        html = a.to_html_table([], ['day'], ['strength', 'position'],
                               cell_format='plot', plot_style='image',
                               rel_plot_dir='./figs', plot_dir=fig_dir,
                               plot_args={'show_colorbar': True})

        expected = '<table><tr><th colspan="2">day</th></tr>'\
            '<tr><th colspan="1">mon</th>'\
            '<th colspan="1">tue</th></tr>' \
            '<tr><td><img src="./figs/xarray_day_mon.png"></td>' \
            '<td><img src="./figs/xarray_day_tue.png"></td>' \
            '</tr></table>'
        self.assert_html_equal(html, expected)

    def assert_html_equal(self, html, expected):
        try:
            from BeautifulSoup import BeautifulSoup
            html, expected = BeautifulSoup(html).prettify(), \
                BeautifulSoup(expected).prettify()
        except ImportError:
            html += '\n'
            expected += '\n'
        if html != expected:
            raise Exception(' html is not as expected. Diff is following '
                            '(1st diff line: html, 2nd diff line: expected)\n%s'
                            % ''.join(difflib.ndiff(html.splitlines(1),
                                                    expected.splitlines(2))))

    def test_table_header(self):
        sh = (2, 3, 4)
        a = xndarray(np.arange(np.prod(sh)).reshape(sh),
                     ['day', 'strength', 'position'],
                     {'day': ['mon', 'tue'],
                      'strength': [0., .5, 1.2],
                      'position': [0, 10, 20, 30]})
        rh, ch = a._html_table_headers(['position'], ['day', 'strength'])

        expected = '<tr><th colspan="2"></th>' \
                   '<th colspan="6">day</th>' \
                   '</tr>' \
                   '<tr>' \
                   '<th colspan="2"></th>' \
                   '<th colspan="3">mon</th>' \
                   '<th colspan="3">tue</th>' \
                   '</tr>' \
                   '<tr>' \
                   '<th colspan="2"></th>' \
                   '<th colspan="6">strength</th>' \
                   '</tr>' \
                   '<tr>' \
                   '<th colspan="2"></th>' \
                   '<th colspan="1">0.0</th>' \
                   '<th colspan="1">0.5</th>' \
                   '<th colspan="1">1.2</th>' \
                   '<th colspan="1">0.0</th>' \
                   '<th colspan="1">0.5</th>' \
                   '<th colspan="1">1.2</th>' \
                   '</tr>'
        self.assert_html_equal(''.join(ch), expected)

        expected = '<tr><th rowspan="4">'\
            '<div class="rotate">position</div></th>' \
                   '<th rowspan="1">0</th>' \
                   '</tr>' \
                   '<tr><th rowspan="1">10</th></tr>' \
                   '<tr><th rowspan="1">20</th></tr>' \
                   '<tr><th rowspan="1">30</th></tr>'
        self.assert_html_equal(''.join([html_row(r) for r in rh]), expected)



# -*- coding: utf-8 -*-

"""Loads and allows configuration of PyHRF."""

from __future__ import with_statement

import os
import os.path as op

from ConfigParser import RawConfigParser, ConfigParser


pyhrf_cfg_path = op.join(os.getenv('HOME'), '.pyhrf/')
if not op.exists(pyhrf_cfg_path):
    os.makedirs(pyhrf_cfg_path)

# Definition of default parameters
DEVEL = 0
ENDUSER = 1
useModesStr = {
    'devel': DEVEL,
    'end_user': ENDUSER,
}
useModes = {useModesStr[k]: k for k in useModesStr}

defaults = {
    'global': {
        'verbosity': 0,
        'use_mode': 'end_user',
        'tmp_prefix': 'pyhrftmp',
        'tmp_path': '/tmp/',
        'spm_path': None,
        'write_texture_minf': False,
        'numpy_floating_point_error': 'warn',
    },
    'log': {
        'level': 'WARNING',
        'log_to_file': False,
    },

    'parallel-LAN': {
        'user': None,
        'remote_host': None,
        'hosts': op.join(pyhrf_cfg_path, 'hosts_LAN'),
        'remote_path': None,
        'niceness': 10,
        'enable_unit_test': 0,
    },
    'parallel-local': {
        'niceness': 10,
        'nb_procs': 1,
    },
    'parallel-cluster': {
        'server': None,
        'server_id': None,
        'user': None,
        'remote_path': None,
        'enable_unit_test': 0,
    },
    'treatment-default': {
        'save_result_dump': 0,
    }
}

section_order = ['global',
                 'log',
                 'parallel-LAN',
                 'parallel-local',
                 'parallel-cluster',
                 'treatment-default']

cfg = defaults.copy()


class ConfigurationError(Exception):
    """Exception class for configuration parsing errors."""
    pass


def write_configuration(cfg_dict, filename, section_order=None):
    if section_order is None:
        section_order = cfg_dict.keys()
    cfgParser = RawConfigParser()
    for section in section_order:
        cfgParser.add_section(section)
        for key, value in cfg_dict[section].iteritems():
            # print 'value:', value
            # print 'str(value):', str(value)
            cfgParser.set(section, key, str(value))

    with open(filename, 'wb') as configfile:
        cfgParser.write(configfile)


def cfg_error_report(cfg, refcfg):
    # Check if all names (section, pname) in cfg are in refcfg
    err = ''
    sections_to_pop = []
    for section in cfg.iterkeys():
        if section not in refcfg.keys():
            sections_to_pop.append(section)
            err += 'Undefined section: "%s"\n' % section
        else:
            labels_to_pop = []
            for pname in cfg[section].iterkeys():
                if pname not in refcfg[section].keys():
                    err += 'Undefined label "%s" (section "%s")\n' \
                        % (pname, section)
                    labels_to_pop.append(pname)
            for l in labels_to_pop:
                cfg[section].pop(l)
    for section in sections_to_pop:
        cfg.pop(section)

    return err


def load_configuration(filename, refcfg, mode='file_only'):
    """
    Load configuration file from 'filename' and check it against 'refcfg'.
    If mode is 'file_only' then only configuration in filename is returned.
    If mode is 'update' then the loaded configuration is updated with 'refcfg' to
    load defaults for unprovided parameters.
    """
    cfgParser = ConfigParser()
    cfgParser.read(filename)
    # Everything is stringified with ConfigParser
    # So cast everything to proper types:
    cfg = {}
    for section in cfgParser.sections():
        cfg[section] = {}
        for pname, pval in cfgParser.items(section):
            cfg[section][pname] = pval
    err = cfg_error_report(cfg, refcfg)
    if err != '':
        msg = 'Errors in configuration file %s\n' % filename
        msg += err
        raise ConfigurationError(msg)

    for section in cfgParser.sections():
        for pname, pval in cfgParser.items(section):
            if isinstance(refcfg[section][pname], float):
                cfg[section][pname] = float(pval)
            elif isinstance(refcfg[section][pname], bool):
                cfg[section][pname] = (pval == 'True')
            elif isinstance(refcfg[section][pname], int):
                cfg[section][pname] = int(pval)
            if pval == 'None':
                cfg[section][pname] = None

    if mode == 'update':
        newcfg = refcfg.copy()
        newcfg.update(cfg)
        return newcfg

    return cfg


# Manage user-defined parameters
# in ~/.pyhrf/config.cfg

user_cfg_fn = op.join(pyhrf_cfg_path, 'config.cfg')
if not op.exists(user_cfg_fn):
    write_configuration(defaults, user_cfg_fn, section_order)
else:
    try:
        cfg = load_configuration(user_cfg_fn, defaults, 'update')
    except ConfigurationError, e:
        print e
    except IOError, e:
        print "Errors reading configuration file:"
        print e
    for section in defaults.iterkeys():
        default_keys = defaults[section].keys()
        unknownLabels = set(default_keys).difference(cfg[section].keys())
        if len(unknownLabels) > 0:
            for l in unknownLabels:
                cfg[section][l] = defaults[section][l]
            write_configuration(cfg, user_cfg_fn, section_order)



# -*- coding: utf-8 -*-
import os, sys
import unittest
import numpy as np

from pyhrf.jde.beta import *
from pyhrf.boldsynth.field import genPotts #, count_homo_cliques
from pyhrf.graph import *
from pyhrf.boldsynth.scenarios import *

from pyhrf.tools import cached_eval

from pyhrf.validation import config
from pyhrf.validation.config import figfn

from pyhrf.boldsynth.pottsfield.swendsenwang import *
from pyhrf.ndarray import stack_cuboids

def dist(x,y):
    return ((x-y)**2).sum()**.5


def beta_estim_obs_field_mc(graph, nbClasses, beta, gridLnz, mcit=1, 
                            cachePotts=False):

    cumulBetaEstim = 0.
    cumul2BetaEstim = 0.

    for mci in xrange(mcit):
        labels = cached_eval(genPotts, (graph, beta, nbClasses), 
                             save=cachePotts, new=(not cachePotts))
        print 'labels:',labels
        betaEstim, _ = beta_estim_obs_field(graph, labels, gridLnz,
                                            method='ML')
        #betaEstim, pb = beta_estim_obs_field(graph, labels, gridLnz,
        #                                     method='MAP')
        print ' -mc- be=%f' %betaEstim
        cumulBetaEstim += betaEstim
        cumul2BetaEstim += betaEstim**2

    meanBetaEstim = cumulBetaEstim / mcit
    stdBetaEstim = ((cumul2BetaEstim / mcit) - meanBetaEstim**2)**.5

    return meanBetaEstim, stdBetaEstim

def test_beta_estim_obs_fields(graphs, betas, nbLabels, pfmethod, mcit=1):
    
    nbGraphs = len(graphs)
    meanBetaEstims = np.zeros(len(betas))
    stdBetaEstims = np.zeros(len(betas))
    stdMCBetaEstims = np.zeros(len(betas))
    gridLnz = range(nbGraphs)
    if mcit == 1:
        cachePotts = True
    else:
        cachePotts = False
        
    # compute partition functions:
    print 'compute partition functions ... nbgraphs:', nbGraphs
    print 'Method =', pfmethod
    print 'nbLabels =', nbLabels
    if pfmethod == 'ES':
        for ig, g in enumerate(graphs):
            grid = cached_eval(Cpt_Vec_Estim_lnZ_Graph_fast3, (g,nbLabels),
                               path=config.cacheDir, prefix='pyhrf_')
            gridLnz[ig] = (grid[0][:-1], grid[1][:-1])    
    elif pfmethod == 'PS':
        for ig, g in enumerate(graphs):
            gridLnz[ig] = cached_eval(Cpt_Vec_Estim_lnZ_Graph, (g,nbLabels),
                                      path=config.cacheDir, new=False,
                                      prefix='pyhrf_')
    elif pfmethod == 'ONSAGER':
        for ig, g in enumerate(graphs):
            betaGrid = cached_eval(Cpt_Vec_Estim_lnZ_Graph,([[]],nbLabels),
                                   path=config.cacheDir)[1]
            lpf = cached_eval(logpf_ising_onsager, (len(g),betaGrid),
                              path=config.cacheDir, prefix='pyhrf_')
            gridLnz[ig] = (lpf, betaGrid)
    else:
        print 'unknown pf method!'
        sys.exit(1)

    print 'gridLnz:'
    print gridLnz

    # Beta estimation:
    print 'beta estimation test loop ...'
    for ib, b in enumerate(betas):
        cumulBetaEstim = 0.
        cumul2BetaEstim = 0.
        cumulStdMC = 0.
        print 'testing with beta =', b, '...'
        for ig, graph in enumerate(graphs):
            mbe, stdmc = beta_estim_obs_field_mc(graph, nbLabels, b,
                                                 gridLnz[ig], mcit, 
                                                 cachePotts=False)
            print '%f(%f)' %(mbe,stdmc)
            cumulBetaEstim += mbe
            cumulStdMC += stdmc
            cumul2BetaEstim += mbe**2
            #print betaEstim, cumulBetaEstim, cumul2BetaEstim

        meanBetaEstims[ib] = cumulBetaEstim / nbGraphs
        stdBetaEstims[ib] = ((cumul2BetaEstim / nbGraphs) - \
                                 meanBetaEstims[ib]**2)**.5
        stdMCBetaEstims[ib] = cumulStdMC / nbGraphs
        #print '-> ', meanBetaEstims[ib], stdBetaEstims[ib]
    return meanBetaEstims, stdBetaEstims, stdMCBetaEstims



class ObsField2DTest(unittest.TestCase):
    """ Test estimation of beta with on observed 2D fields
    """

    def setUp(self):
        self.outDir = os.path.join(config.plotSaveDir, 
                                   './BetaEstimation')
        if not os.path.exists(self.outDir):
            os.makedirs(self.outDir)

        self.plot = config.savePlots

        shape = (15,15)
        mask = np.ones(shape, dtype=int) #full mask
        self.g = graph_from_lattice(mask, kerMask=kerMask2D_4n, toroidal=True)
        
    def test_single_PFES_MAP(self):
        """PF estimation method : extrapolation scheme. MAP on p(beta|label).
        """
        # generate a field:
        beta = 0.4
        nbClasses = 2
	nbTests = 30
	#print 'g:', self.g
        
	# partition function estimation
        gridLnz = Cpt_Vec_Estim_lnZ_Graph_fast3(self.g, nbClasses)
        gridPace = gridLnz[1][1] - gridLnz[1][0]
	#print 'generating potts ..., beta =', beta
	bes = np.zeros(nbTests)
        for t in xrange(nbTests):
	    labels = genPotts(self.g, beta, nbClasses)
        
	    # beta estimation
	    be, pb = beta_estim_obs_field(self.g, labels, gridLnz, 'MAP')
	    print 'betaMAP:', be
	    bes[t] = be
	print 'mean betaMAP:', bes.mean(), bes.std()
        assert abs(bes.mean()-beta) <= gridPace


    def test_single_PFES_ML(self):
        """PF estimation method : extrapolation scheme. ML on p(label|beta).
        """
        # generate a field:
        beta = 0.4
        nbClasses = 2
        print 'generating potts ..., beta =', beta
        labels = genPotts(self.g, beta, nbClasses)
        # partition function estimation
        gridLnz = Cpt_Vec_Estim_lnZ_Graph_fast3(self.g, nbClasses)
        gridPace = gridLnz[1][1] - gridLnz[1][0]
        # beta estimation
        be, pb = beta_estim_obs_field(self.g, labels, gridLnz, 'ML')
        print 'betaML:', be
        assert abs(be-beta) <= gridPace


    def test_single_PFPS_MAP(self):
        """PF estimation method : path sampling. MAP on p(beta|label).
        """
        # generate a field:
        beta = 0.4
        nbClasses = 2
        print 'generating potts ..., beta =', beta
        labels = genPotts(self.g, beta, nbClasses)
        # partition function estimation
        gridLnz = Cpt_Vec_Estim_lnZ_Graph(self.g, nbClasses)
        gridPace = gridLnz[1][1] - gridLnz[1][0]
        # beta estimation
        be, pb = beta_estim_obs_field(self.g, labels, gridLnz, 'MAP')
        print 'betaMAP:', be
        assert abs(be-beta) <= gridPace


    def test_single_PFPS_ML(self):
        """PF estimation method : path sampling. ML on p(label|beta).
        """
        # generate a field:
        beta = 0.4
        nbClasses = 2
        print 'generating potts ..., beta =', beta
        labels = genPotts(self.g, beta, nbClasses)
        # partition function estimation
        gridLnz = Cpt_Vec_Estim_lnZ_Graph(self.g, nbClasses)
        gridPace = gridLnz[1][1] - gridLnz[1][0]
        # beta estimation
        be, pb = beta_estim_obs_field(self.g, labels, gridLnz, 'ML')
        print 'betaML:', be
        assert abs(be-beta) <= gridPace


    def test_single_surface_PFPS_ML(self):
        """PF estimation method : path sampling. ML on p(label|beta).
        topology from a surfacic RDI
        """
        # generate a field:
        beta = 0.4
        nbClasses = 2
        print 'generating potts ..., beta =', beta

        
        # grab surfacic data:
        from pyhrf.graph import graph_from_mesh, sub_graph, graph_is_sane
        from pyhrf.tools._io.tio import Texture
        from soma import aims
        print 'import done'
        roiId = 20
        mfn = pyhrf.get_data_file_name('right_hemisphere.mesh')
        print 'mesh file:', mfn
        mesh = aims.read(mfn)
        print 'mesh read'
        triangles = [t.arraydata() for t in mesh.polygon().list()]
        print 'building graph ... '
        wholeGraph = graph_from_mesh(triangles)
        
        roiMaskFile = pyhrf.get_data_file_name('roimask_gyrii_tuned.tex')
        roiMask = Texture.read(roiMaskFile).data.astype(int)
        mroi = np.where(roiMask==roiId)
        g, nm = sub_graph(wholeGraph, mroi[0])
        print "g:", len(g), len(g[0])

        nnodes = len(g)
        points = np.vstack([v.arraydata() for v in mesh.vertex().list()])
        weights = [[1./dist(points[j],points[k]) for k in g[j]]
                   for j in xrange(nnodes)]
        print "weights:", len(weights), len(weights[0])
        
        if 1:
            for j in xrange(nnodes):
                s = sum(weights[j]) * 1.
                for k in xrange(len(weights[j])):
                    weights[j][k] = weights[j][k]/s * len(weights[j])

        labels = genPotts(g, beta, nbClasses, weights=weights)
        print labels
        # partition function estimation
        gridLnz = Cpt_Vec_Estim_lnZ_Graph(g, nbClasses, 
                                          GraphWeight=weights)

        print 'gridLnz with weights:'
        print gridLnz

        # beta estimation
        be, pb = beta_estim_obs_field(g, labels, gridLnz, 'ML', weights)
        print 'betaML:', be

        weights = None
        gridLnz = Cpt_Vec_Estim_lnZ_Graph(g, nbClasses, 
                                          GraphWeight=weights)

        
        print 'gridLnz without weights:'
        print gridLnz

        # beta estimation
        be, pb = beta_estim_obs_field(g, labels, gridLnz, 'ML', weights)
        print 'betaML:', be

        gridPace = gridLnz[1][1] - gridLnz[1][0]
        assert abs(be-beta) <= gridPace


    def test_single_Onsager_ML(self):
        """PF method: Onsager. ML on p(beta|label).
        """
        # generate a field:
        beta = 0.4
        nbClasses = 2
        print 'generating potts ..., beta =', beta
        labels = genPotts(self.g, beta, nbClasses)
        # partition function estimation
        dbeta = 0.05
        betaGrid = np.arange(0, 1.5, dbeta)
        lpf = logpf_ising_onsager(labels.size, betaGrid)
        #lpf, betaGrid = Cpt_Vec_Estim_lnZ_Graph(g, nbClasses)
        print 'lpf:'
        print lpf
        betaML = beta_estim_obs_field(self.g, labels, (lpf, betaGrid), 'ML')
        print 'betaML:', betaML

    def test_single_Onsager_MAP(self):
        """PF method: Onsager. MAP on p(label|beta).
        """
        # generate a field:
        beta = 0.4
        nbClasses = 2
        print 'generating potts ..., beta =', beta
        labels = genPotts(self.g, beta, nbClasses)
        # partition function estimation
        dbeta = 0.05
        betaGrid = np.arange(0, 1.5, dbeta)
        lpf = logpf_ising_onsager(labels.size, betaGrid)
        #lpf, betaGrid = Cpt_Vec_Estim_lnZ_Graph(g, nbClasses)
        print 'lpf:'
        print lpf
        betaMAP = beta_estim_obs_field(self.g, labels, (lpf, betaGrid), 'MAP')
        print 'betaMAP:', betaMAP


    def test_MC_comp_pfmethods_ML_10x10(self):
        self.MC_comp_pfmethods_ML((10,10))

    def test_MC_comp_pfmethods_ML_30x30(self):
        self.MC_comp_pfmethods_ML((30,30))

    def test_MC_comp_pfmethods_ML_3C_50x50(self):
        self.MC_comp_pfmethods_ML_3C((50,50))

    def test_MC_comp_pfmethods_ML_3C_10x10(self):
        self.MC_comp_pfmethods_ML_3C((10,10))

    def test_MC_comp_pfmethods_ML_3C_20x20(self):
        self.MC_comp_pfmethods_ML_3C((20,20))

    def test_MC_comp_pfmethods_ML_100x100(self):
        self.MC_comp_pfmethods_ML((100, 100))
        
    def test_MC_comp_pfmethods_ML_3C_30x30(self):
        self.MC_comp_pfmethods_ML_3C((30, 30))        

    def MC_comp_pfmethods_ML(self, shape):

        mask = np.ones(shape, dtype=int) #full mask
        g = graph_from_lattice(mask, kerMask=kerMask2D_4n, toroidal=True)

        # Betas to iterate over ...
        betaMax = 1.4
        dbeta = .1
        betas = np.arange(0., betaMax, dbeta)

        # nb of monte carlo iterations
        mcit = 40#100
        
        # nb of classes
        nbc = 2

        # Path sampling:
        mBePS, stdBePS, stdBeMCPS = cached_eval(test_beta_estim_obs_fields,
                                                ([g], betas, nbc, 'PS',
                                                 mcit), path=config.cacheDir,
                                                prefix='pyhrf_',
                                                new=False)
        
        # Extrapolation scheme:
        mBeES, stdBeES, stdBeMCES = cached_eval(test_beta_estim_obs_fields,
                                                ([g], betas, nbc, 'ES',
                                                 mcit), path=config.cacheDir,
                                                prefix='pyhrf_',
                                                new=False)

        # Onsager:
        #mBeON, stdBeON, stdBeMCON = cached_eval(test_beta_estim_obs_fields,
        #                                        ([g], betas, nbc, 'ONSAGER',
        #                                         mcit), path=config.cacheDir,
        #                                        prefix='pyhrf_',
        #                                        new=True)
        
        if self.plot:
            import matplotlib.pyplot as plt
            plt.figure()

            plt.errorbar(betas, mBeES, stdBeMCES, fmt='r', label='ES')
            plt.errorbar(betas, mBePS, stdBeMCPS, fmt='b', label='PS')
            #plt.errorbar(betas, mBeON, stdBeMCON, fmt='g', label='Onsager')
            plt.plot(betas, betas, 'k', label='true')

            #plt.xlabel('simulated beta')
            #plt.ylabel('beta MAP')
            plt.legend(loc='upper left')
            plt.ylim((0.0,1.45))
            #plt.title('MC validation (%d it) for beta estimation on '\
            #                ' observed fields.' %mcit)
            #fn = 'comp_obs_field_MC_PS_ES_Onsager_%dx%d' %shape 
            fn = 'comp_obs_field_MC_PS_%dx%d' %shape 
            figFn = os.path.join(self.outDir, figfn(fn))
            plt.savefig(figFn)



    def MC_comp_pfmethods_ML_3C(self, shape):

        mask = np.ones(shape, dtype=int) #full mask
        g = graph_from_lattice(mask, kerMask=kerMask2D_4n, toroidal=True)

        # Betas to iterate over ...
        betaMax = 1.4
        dbeta = .1
        betas = np.arange(0., betaMax, dbeta)

        # nb of monte carlo iterations
        mcit = 100#100
        
        # nb of classes
        nbc = 3

        # Path sampling:
        mBePS, stdBePS, stdBeMCPS = cached_eval(test_beta_estim_obs_fields,
                                                ([g], betas, nbc, 'PS',
                                                 mcit), path=config.cacheDir,
                                                prefix='pyhrf_',
                                                new=False)
        
        # Extrapolation scheme:
        mBeES, stdBeES, stdBeMCES = cached_eval(test_beta_estim_obs_fields,
                                                ([g], betas, nbc, 'ES',
                                                 mcit), path=config.cacheDir,
                                                prefix='pyhrf_',
                                                new=False)

        # Onsager:
        #mBeON, stdBeON, stdBeMCON = cached_eval(test_beta_estim_obs_fields,
        #                                        ([g], betas, nbc, 'ONSAGER',
        #                                         mcit), path=config.cacheDir,
        #                                        prefix='pyhrf_',
        #                                        new=True)
        
        if self.plot:
            import plt
            plt.figure()

            plt.errorbar(betas, mBeES, stdBeMCES, fmt='r', label='ES')
            plt.errorbar(betas, mBePS, stdBeMCPS, fmt='b', label='PS')
            #plt.errorbar(betas, mBeON, stdBeMCON, fmt='g', label='Onsager')
            plt.plot(betas, betas, 'k', label='true')
            #plt.xlabel('simulated beta')
            #plt.ylabel('beta MAP')
            plt.legend(loc='upper left')
            #plt.title('MC validation (%d it) for beta estimation on '\
            #                ' observed fields.' %mcit)
            fn = 'comp_obs_field_3class_MC_PS_ES_%dx%d' %shape 
            figFn = os.path.join(self.outDir, figfn(fn))
            plt.savefig(figFn)
            
            
            plt.figure()
            grid = cached_eval(Cpt_Vec_Estim_lnZ_Graph_fast3, (g,nbc),
                               path=config.cacheDir, prefix='pyhrf_')
            lnzES, betas = (grid[0][:-1], grid[1][:-1])    
            lnzPS, betas = cached_eval(Cpt_Vec_Estim_lnZ_Graph, (g,nbc),
                                       path=config.cacheDir, new=False,
                                       prefix='pyhrf_')

            plt.plot(betas, lnzES, 'r', label='ES')
            plt.plot(betas, lnzPS, 'b', label='PS')
            fn = 'comp_PF_3class_PS_ES_%dx%d' %shape 
            figFn = os.path.join(self.outDir, figfn(fn))
            plt.savefig(figFn)
